"""Test for the model mapper."""
import json
import unittest
from datetime import date, datetime, timedelta

from pymultimatic.model import ActiveFunction, OperatingModes, QuickModes, mapper
from tests.conftest import path, senso_responses_files_paths


class MapperTest(unittest.TestCase):
    """Test class."""

    def test_map_quick_mode(self) -> None:
        with open(path("files/responses/quick_mode"), "r") as file:
            json_raw = json.loads(file.read())
            q_m = mapper.map_quick_mode(json_raw)
            self.assertEqual(QuickModes.SYSTEM_OFF, q_m)
            self.assertEqual(0, q_m.duration)

    def test_map_quick_mode_no_duration(self) -> None:
        with open(path("files/responses/quick_mode_no_duration"), "r") as file:
            json_raw = json.loads(file.read())
            q_m = mapper.map_quick_mode(json_raw)
            self.assertEqual(QuickModes.SYSTEM_OFF, q_m)
            self.assertIsNone(q_m.duration)

    def test_map_holiday(self) -> None:
        with open(path("files/responses/holiday_mode"), "r") as file:
            json_raw = json.loads(file.read())
            holiday = mapper.map_holiday_mode(json_raw)
            self.assertEqual(False, holiday.is_active)

    def test_map_zones(self) -> None:
        with open(path("files/responses/zones"), "r") as file:
            json_raw = json.loads(file.read())
            zones = mapper.map_zones(json_raw)
            self.assertEqual(2, len(zones))

    def test_map_zones_senso_vr920(self) -> None:
        with open(path("files/responses/senso/vr920/zones"), "r") as file:
            json_raw = json.loads(file.read())
            zones = mapper.map_zones(json_raw)
            self.assertEqual(1, len(zones))

    def test_map_zones_senso_vr921(self) -> None:
        with open(path("files/responses/senso/vr921/zones"), "r") as file:
            json_raw = json.loads(file.read())
            zones = mapper.map_zones(json_raw)
            self.assertEqual(3, len(zones))

    def test_map_zones_3_zones(self) -> None:
        with open(path("files/responses/zones_3_zones"), "r") as file:
            json_raw = json.loads(file.read())
            zones = mapper.map_zones(json_raw)
            self.assertEqual(3, len(zones))

    def test_map_zones_quick_veto_no_heating_config(self) -> None:
        with open(path("files/responses/zones_missing_heating_config_quick_veto"), "r") as file:
            json_raw = json.loads(file.read())
            zones = mapper.map_zones(json_raw)
            self.assertEqual(3, len(zones))

    def test_map_dhw(self) -> None:
        with open(path("files/responses/dhws"), "r") as file:
            raw_dhw = json.loads(file.read())
            dhw = mapper.map_dhw(raw_dhw)
            self.assertIsNotNone(dhw.hotwater)
            self.assertIsNone(dhw.hotwater.temperature)
            self.assertIsNotNone(dhw.circulation)

    def test_map_dhw_senso(self) -> None:
        for file_path in senso_responses_files_paths("dhws"):
            with open(file_path, "r") as file:
                raw_dhw = json.loads(file.read())
                dhw = mapper.map_dhw(raw_dhw)
                self.assertIsNotNone(dhw.hotwater)
                self.assertIsNone(dhw.hotwater.temperature)
                self.assertIsNotNone(dhw.hotwater.target_high)
                self.assertIsNotNone(dhw.circulation)
                self.assertIsNotNone(dhw.hotwater.active_mode)

    def test_map_dhw_no_timeprogram(self) -> None:
        with open(path("files/responses/dhws_minimal"), "r") as file:
            raw_dhw = json.loads(file.read())
            dhw = mapper.map_dhw(raw_dhw)
            self.assertIsNotNone(dhw.hotwater)
            self.assertIsNotNone(dhw.circulation)
            self.assertIsNotNone(dhw.circulation.time_program)
            self.assertIsNotNone(dhw.hotwater.time_program)

    def test_map_dhw_no_timeprogram_senso(self) -> None:
        for file_path in senso_responses_files_paths("dhw"):
            with open(file_path, "r") as file:
                raw_dhw = json.loads(file.read())
                dhw = mapper.map_dhw(raw_dhw)
                self.assertIsNotNone(dhw.hotwater)
                self.assertIsNotNone(dhw.circulation)
                self.assertIsNotNone(dhw.circulation.time_program)
                self.assertIsNotNone(dhw.hotwater.time_program)

    def test_map_dhw_empty_timeprogram(self) -> None:
        with open(path("files/responses/dhws_empty_timprogram"), "r") as file:
            raw_dhw = json.loads(file.read())
            dhw = mapper.map_dhw(raw_dhw)
            self.assertIsNotNone(dhw.hotwater)
            self.assertIsNotNone(dhw.circulation)
            self.assertIsNotNone(dhw.circulation.time_program)
            self.assertIsNotNone(dhw.hotwater.time_program)

    def test_map_zone_cooling(self) -> None:
        """Test map zone with cooling."""
        with open(path("files/responses/systemcontrol_ventilation"), "r") as file:
            system = json.loads(file.read())
        zones = mapper.map_zones_from_system(system)
        self.assertIsNotNone(zones)

    def test_map_zone_empty(self) -> None:
        """Test map no zone."""
        zone = mapper.map_zone({})
        self.assertIsNone(zone)

    def test_map_zone_no_active_function(self) -> None:
        """Test map a zone without active function"""
        with open(path("files/responses/zone_no_active_function"), "r") as file:
            zone_file = json.loads(file.read())

        zone = mapper.map_zone(zone_file)
        self.assertEqual(ActiveFunction.STANDBY, zone.active_function)

    def test_map_quick_mode_from_system(self) -> None:
        """Test map quick mode."""
        with open(path("files/responses/systemcontrol_hotwater_boost"), "r") as file:
            system = json.loads(file.read())

        quick_mode = mapper.map_quick_mode_from_system(system)
        self.assertEqual(QuickModes.HOTWATER_BOOST.name, quick_mode.name)

    def test_map_quick_mode_quick_veto(self) -> None:
        """Test map quick veto."""
        with open(path("files/responses/systemcontrol_quick_veto"), "r") as file:
            system = json.loads(file.read())

        quick_mode = mapper.map_quick_mode_from_system(system)
        self.assertIsNone(quick_mode)

    def test_map_quick_veto_zone(self) -> None:
        """Test map quick veto zone."""
        with open(path("files/responses/systemcontrol_quick_veto"), "r") as file:
            system = json.loads(file.read())

        zones = mapper.map_zones_from_system(system)

        for zone in zones:
            if zone.id == "Control_ZO2":
                self.assertIsNotNone(zone.quick_veto)
                self.assertEqual(18.5, zone.quick_veto.target)
                self.assertIsNone(zone.quick_veto.duration)
                return
        self.fail("Wrong zone found")

    def test_map_quick_veto_senso_zone(self) -> None:
        """Test map quick veto zone for Senso."""
        with open(path("files/responses/senso/vr921/systemcontrol_quick_veto"), "r") as file:
            system = json.loads(file.read())
        # update of the expiry date for the tests
        system.get("body", {}).get("zones", [])[0]["configuration"]["quick_veto"]["expires_at"] = (
            datetime.utcnow() + timedelta(hours=2)
        ).strftime(mapper._DATE_TIME_FORMAT)
        zones = mapper.map_zones_from_system(system)

        for zone in zones:
            if zone.id == "Control_ZO1":
                self.assertIsNotNone(zone.quick_veto)
                self.assertEqual(22.0, zone.quick_veto.target)
                self.assertTrue(zone.quick_veto.duration <= 120)
                return
        self.fail("Wrong zone found")

    def test_map_quick_veto_senso_zone_after_put(self) -> None:
        """Test map quick veto zone for Senso after put quick veto.
        The expiry date is not immediately filled in."""
        with open(
            path("files/responses/senso/vr921/systemcontrol_quick_veto_after_put"), "r"
        ) as file:
            system = json.loads(file.read())
        zones = mapper.map_zones_from_system(system)

        for zone in zones:
            if zone.id == "Control_ZO1":
                self.assertIsNotNone(zone.quick_veto)
                self.assertEqual(22.0, zone.quick_veto.target)
                self.assertIsNone(zone.quick_veto.duration)
                return
        self.fail("Wrong zone found")

    def test_map_no_quick_mode(self) -> None:
        """Test map no quick mode."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            system = json.loads(file.read())

        quick_mode = mapper.map_quick_mode_from_system(system)
        self.assertIsNone(quick_mode)

    def test_map_outdoor_temp(self) -> None:
        """Test map outdoor temperature."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            system = json.loads(file.read())

        temp = mapper.map_outdoor_temp_from_system(system)
        self.assertEqual(6.3, temp)

    def test_map_no_outdoor_temp(self) -> None:
        """Test map no outdoor temperature."""
        with open(path("files/responses/systemcontrol_no_outside_temp"), "r") as file:
            system = json.loads(file.read())

        temp = mapper.map_outdoor_temp_from_system(system)
        self.assertIsNone(temp)

    def test_rooms_none(self) -> None:
        """Test map no rooms."""
        rooms = mapper.map_rooms(None)
        self.assertIsNotNone(rooms)
        self.assertEqual(0, len(rooms))

    def test_rooms_empty(self) -> None:
        """Test map empty rooms."""
        rooms = mapper.map_rooms({})
        self.assertIsNotNone(rooms)
        self.assertEqual(0, len(rooms))

    def test_rooms_correct(self) -> None:
        """Test map rooms."""
        with open(path("files/responses/rooms"), "r") as file:
            raw_rooms = json.loads(file.read())

        rooms = mapper.map_rooms(raw_rooms)
        self.assertIsNotNone(rooms)
        self.assertEqual(4, len(rooms))

        room0 = rooms[0]

        self.assertEqual(0, room0.id)
        self.assertEqual("Room 1", room0.name)
        self.assertEqual(OperatingModes.AUTO, room0.operating_mode)
        self.assertEqual(False, room0.window_open)
        self.assertEqual(17.5, room0.target_high)
        self.assertEqual(17.9, room0.temperature)
        self.assertIsNone(room0.quick_veto)
        self.assertEqual(False, room0.child_lock)

    def test_room_empty(self) -> None:
        """Test map empty room."""
        rooms = mapper.map_room({})
        self.assertIsNone(rooms)

    def test_room_quick_veto(self) -> None:
        """Test map quick veto room."""
        with open(path("files/responses/rooms_quick_veto"), "r") as file:
            raw_rooms = json.loads(file.read())

        rooms = mapper.map_rooms(raw_rooms)
        self.assertIsNotNone(rooms)
        self.assertEqual(4, len(rooms))

        room0 = rooms[0]

        self.assertEqual(0, room0.id)
        self.assertEqual("Room 1", room0.name)
        self.assertEqual(OperatingModes.AUTO, room0.operating_mode)
        self.assertEqual(False, room0.window_open)
        self.assertEqual(20.0, room0.target_high)
        self.assertEqual(17.9, room0.temperature)
        self.assertIsNotNone(room0.quick_veto)
        self.assertEqual(20.0, room0.quick_veto.target)
        self.assertEqual(False, room0.child_lock)

    def test_map_devices(self) -> None:
        """Test map devices."""
        with open(path("files/responses/rooms"), "r") as file:
            raw_rooms = json.loads(file.read())

        rooms = mapper.map_rooms(raw_rooms)
        self.assertIsNotNone(rooms)
        self.assertEqual(4, len(rooms))

        devices_room0 = rooms[0].devices
        devices_room1 = rooms[1].devices

        self.assertIsNotNone(devices_room0)
        self.assertEqual(1, len(devices_room0))
        self.assertIsNotNone(devices_room1)
        self.assertEqual(2, len(devices_room1))

        self.assertEqual("Device 1", devices_room0[0].name)
        self.assertEqual("R13456789012345678901234", devices_room0[0].sgtin)
        self.assertEqual("VALVE", devices_room0[0].device_type)
        self.assertEqual(True, devices_room0[0].radio_out_of_reach)
        self.assertEqual(True, devices_room0[0].radio_out_of_reach)

        self.assertEqual("Device 1", devices_room1[0].name)
        self.assertEqual("R20123456789012345678900", devices_room1[0].sgtin)
        self.assertEqual("VALVE", devices_room1[0].device_type)
        self.assertEqual(False, devices_room1[0].radio_out_of_reach)
        self.assertEqual(False, devices_room1[0].radio_out_of_reach)

        self.assertEqual("Device 2", devices_room1[1].name)
        self.assertEqual("R20123456789012345678999", devices_room1[1].sgtin)
        self.assertEqual("VALVE", devices_room1[1].device_type)
        self.assertEqual(False, devices_room1[1].radio_out_of_reach)
        self.assertEqual(False, devices_room1[1].radio_out_of_reach)

    def test_map_devices_no_name(self) -> None:
        """Test map devices."""
        with open(path("files/responses/room_empty_device_name"), "r") as file:
            raw_room = json.loads(file.read())

        room = mapper.map_room(raw_room)
        self.assertIsNotNone(room)
        self.assertEqual(2, len(room.devices))

        self.assertEqual("", room.devices[0].name)
        self.assertEqual(None, room.devices[1].name)

    def test_holiday_mode_none(self) -> None:
        """Test map no holiday mode."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            raw_system = json.loads(file.read())

        holiday_mode = mapper.map_holiday_mode_from_system(raw_system)
        self.assertIsNotNone(holiday_mode)
        self.assertFalse(holiday_mode.is_active)
        self.assertIsNotNone(holiday_mode.start_date)
        self.assertIsNotNone(holiday_mode.end_date)
        self.assertIsNotNone(holiday_mode.target)
        self.assertFalse(holiday_mode.is_applied)
        self.assertIsNone(holiday_mode.active_mode)

    def test_holiday_mode(self) -> None:
        """Test map holiday mode."""
        with open(path("files/responses/systemcontrol_holiday"), "r") as file:
            raw_system = json.loads(file.read())

        holiday_mode = mapper.map_holiday_mode_from_system(raw_system)
        quick_mode = mapper.map_quick_mode_from_system(raw_system)
        self.assertEqual(QuickModes.HOLIDAY, quick_mode)
        self.assertIsNotNone(holiday_mode)
        self.assertTrue(holiday_mode.is_active)
        self.assertEqual(date(2019, 1, 2), holiday_mode.start_date)
        self.assertEqual(date(2019, 1, 3), holiday_mode.end_date)
        self.assertEqual(15, holiday_mode.target)

    def test_map_circulation(self) -> None:
        """Test map circulation."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            raw_system = json.loads(file.read())

        circulation = mapper.map_circulation_from_system(raw_system)
        self.assertEqual(OperatingModes.AUTO, circulation.operating_mode)
        self.assertEqual("Control_DHW", circulation.id)
        self.assertIsNone(circulation.temperature)
        self.assertIsNone(circulation.target_high)
        self.assertIsNotNone(circulation.time_program)
        self.assertIsNotNone(circulation.time_program.days["monday"].settings[0].setting)

    def test_hot_water(self) -> None:
        """Test map hot water."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            raw_system = json.loads(file.read())
        with open(path("files/responses/livereport"), "r") as file:
            raw_livereport = json.loads(file.read())

        hot_water = mapper.map_hot_water_from_system(raw_system, raw_livereport)
        self.assertEqual(44.5, hot_water.temperature)
        self.assertEqual(51, hot_water.target_high)
        self.assertEqual(OperatingModes.AUTO, hot_water.operating_mode)
        self.assertEqual("Control_DHW", hot_water.id)
        self.assertIsNotNone(hot_water.time_program.days["monday"].settings[0].setting)

    def test_no_hotwater(self) -> None:
        """Test map no hot water."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            raw_system = json.loads(file.read())

        raw_system["body"]["dhw"] = []

        hot_water = mapper.map_hot_water_from_system(raw_system, {})
        self.assertIsNone(hot_water)

    def test_hot_water_no_current_temp(self) -> None:
        """Test map hot water no live report."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            raw_system = json.loads(file.read())

        hot_water = mapper.map_hot_water_from_system(raw_system, json.loads("{}"))
        self.assertEqual(None, hot_water.temperature)
        self.assertEqual(51, hot_water.target_high)
        self.assertEqual(OperatingModes.AUTO, hot_water.operating_mode)
        self.assertEqual("Control_DHW", hot_water.id)

    def test_boiler_status(self) -> None:
        """Test map boiler status."""
        with open(path("files/responses/hvacstate"), "r") as file:
            hvac = json.loads(file.read())

        hvac_status = mapper.map_hvac_status(hvac)
        boiler_status = hvac_status.boiler_status
        self.assertEqual("...", boiler_status.hint)
        self.assertEqual("...", boiler_status.description)
        self.assertEqual("S.8", boiler_status.status_code)
        self.assertEqual(
            "Mode chauffage : Arrêt temporaire après une " "opération de chauffage",
            hvac_status.boiler_status.title,
        )
        self.assertEqual("VC BE 246/5-3", hvac_status.boiler_status.device_name)
        self.assertFalse(hvac_status.boiler_status.is_error)
        self.assertEqual(datetime.fromtimestamp(1545896904282 / 1000), boiler_status.timestamp)

    def test_boiler_status_no_live_report(self) -> None:
        """Test map boiler status no live report."""
        with open(path("files/responses/hvacstate"), "r") as file:
            hvac = json.loads(file.read())

        hvac_status = mapper.map_hvac_status(hvac)
        self.assertEqual("...", hvac_status.boiler_status.hint)
        self.assertEqual("...", hvac_status.boiler_status.description)
        self.assertEqual("S.8", hvac_status.boiler_status.status_code)
        self.assertEqual(
            "Mode chauffage : Arrêt temporaire après une " "opération de chauffage",
            hvac_status.boiler_status.title,
        )
        self.assertEqual("VC BE 246/5-3", hvac_status.boiler_status.device_name)
        self.assertFalse(hvac_status.boiler_status.is_error)
        self.assertEqual(
            datetime.fromtimestamp(1545896904282 / 1000),
            hvac_status.boiler_status.timestamp,
        )

    def test_boiler_status_empty(self) -> None:
        """Test map empty boiler status."""
        with open(path("files/responses/hvacstate_empty"), "r") as file:
            hvac = json.loads(file.read())

        hvac_status = mapper.map_hvac_status(hvac)
        self.assertIsNone(hvac_status.boiler_status)

    def test_hot_water_alone(self) -> None:
        """Test map hot water."""
        with open(path("files/responses/hotwater"), "r") as file:
            raw_hotwater = json.loads(file.read())

        hotwater = mapper.map_hot_water(raw_hotwater, "control_dhw")
        self.assertEqual("control_dhw", hotwater.id)
        self.assertEqual(OperatingModes.AUTO, hotwater.operating_mode)
        self.assertIsNotNone(hotwater.time_program.days["monday"].settings[0].setting)

    def test_hot_water_alone_none(self) -> None:
        """Test map hot water."""
        hotwater = mapper.map_hot_water(None, "control_dhw")
        self.assertIsNone(hotwater)

    def test_circulation_alone(self) -> None:
        """Test map circulation."""
        with open(path("files/responses/circulation"), "r") as file:
            raw_circulation = json.loads(file.read())

        circulation = mapper.map_circulation_alone(raw_circulation, "control_dhw")
        self.assertEqual("control_dhw", circulation.id)
        self.assertEqual(OperatingModes.AUTO, circulation.operating_mode)
        self.assertIsNotNone(circulation.time_program)
        self.assertIsNotNone(circulation.time_program.days["monday"].settings[0].setting)

    def test_circulation_alone_none(self) -> None:
        """Test map circulation."""
        circulation = mapper.map_circulation_alone(None, "control_dhw")
        self.assertIsNone(circulation)

    def test_no_circulation(self) -> None:
        """Test map no circulation."""
        with open(path("files/responses/systemcontrol"), "r") as file:
            raw_system = json.loads(file.read())

        raw_system["body"]["dhw"] = []

        circulation = mapper.map_circulation_from_system(raw_system)
        self.assertIsNone(circulation)

    def test_errors_no_error(self) -> None:
        """Test map no errors."""
        with open(path("files/responses/hvacstate"), "r") as file:
            raw_hvac = json.loads(file.read())

        errors = mapper.map_errors(raw_hvac)
        self.assertEqual(0, len(errors))

    def test_errors_with_errors(self) -> None:
        """Test map hvac errors."""
        with open(path("files/responses/hvacstate_errors"), "r") as file:
            raw_hvac = json.loads(file.read())

        errors = mapper.map_errors(raw_hvac)
        self.assertEqual(1, len(errors))
        self.assertEqual(mapper._datetime(1562909693021), errors[0].timestamp)
        self.assertEqual("...", errors[0].description)
        self.assertEqual("Défaut : Bus de communication eBus", errors[0].title)
        self.assertEqual("VR920", errors[0].device_name)
        self.assertEqual("F.900", errors[0].status_code)

    def test_map_facility_detail(self) -> None:
        with open(path("files/responses/facilities"), "r") as file:
            facilities = json.loads(file.read())

        sys_info = mapper.map_facility_detail(facilities)
        self.assertEqual("1234567890123456789012345678", sys_info.serial_number)
        self.assertEqual("Home", sys_info.name)
        self.assertEqual("01:23:45:67:89:AB", sys_info.ethernet_mac)
        self.assertEqual("1.2.3", sys_info.firmware_version)

    def test_map_facility_detail_senso(self) -> None:
        for file_path in senso_responses_files_paths("facilities_list"):
            with open(file_path, "r") as file:
                facilities = json.loads(file.read())

            sys_info = mapper.map_facility_detail(facilities)
            self.assertEqual("SERIAL_NUMBER", sys_info.serial_number)
            self.assertEqual("Maison", sys_info.name)
            self.assertEqual("01:23:45:67:89:AB", sys_info.ethernet_mac)
            self.assertEqual("0357.27.06", sys_info.firmware_version)

    def test_map_system_info_specific_serial(self) -> None:
        with open(path("files/responses/facilities_multiple"), "r") as file:
            facilities = json.loads(file.read())

        sys_info = mapper.map_facility_detail(facilities, "888")
        self.assertEqual("888", sys_info.serial_number)
        self.assertEqual("Home2", sys_info.name)
        self.assertEqual("6.6.6", sys_info.firmware_version)

    def test_map_hvac_sync_state_none(self) -> None:
        self.assertIsNone(mapper.map_hvac_sync_state(None))

    def test_map_reports(self) -> None:
        with open(path("files/responses/livereport"), "r") as file:
            livereport = json.loads(file.read())
        reports = mapper.map_reports(livereport)
        self.assertEqual(5, len(reports))
        self.assertEqual("VRC700 MultiMatic", reports[0].device_name)
        self.assertEqual("Control_SYS_MultiMatic", reports[0].device_id)
        self.assertEqual("bar", reports[0].unit)
        self.assertEqual(1.9, reports[0].value)
        self.assertEqual("Water pressure", reports[0].name)
        self.assertEqual("WaterPressureSensor", reports[0].id)

    def test_map_reports_no_livereport(self) -> None:
        reports = mapper.map_reports(None)
        self.assertEqual(0, len(reports))

    def test_map_ventilation(self) -> None:
        with open(path("files/responses/systemcontrol_ventilation"), "r") as file:
            system = json.loads(file.read())

        ventilation = mapper.map_ventilation_from_system(system)
        self.assertIsNotNone(ventilation)
        self.assertEqual("_template", ventilation.id)
        self.assertEqual("Ventilation", ventilation.name)
        self.assertEqual(OperatingModes.AUTO, ventilation.operating_mode)
        self.assertEqual(3, ventilation.target_high)
        self.assertEqual(1, ventilation.target_low)
        self.assertIsNone(ventilation.temperature)

    def test_map_zone_quickveto(self) -> None:
        with open(path("files/responses/zone_no_quickveto"), "r") as file:
            raw_zone = json.loads(file.read())
            zone = mapper.map_zone(raw_zone)
            self.assertIsNotNone(zone)
            self.assertIsNone(zone.quick_veto)

    def test_map_system_no_config_rbr(self) -> None:
        with open(path("files/responses/systemcontrol_zone_no_config_rbr"), "r") as file:
            raw_system = json.loads(file.read())
            zones = mapper.map_zones_from_system(raw_system)
            self.assertIsNotNone(zones)
            self.assertIsNotNone(zones[0])
            self.assertIsNotNone(zones[1])

    def test_map_emf_reports(self) -> None:
        with open(path("files/responses/emf_devices"), "r") as file:
            raw_emf_reports = json.loads(file.read())
        reports = mapper.map_emf_reports(raw_emf_reports)
        self.assertEqual(7, len(reports))
        self.assertEqual("VWF 117/4", reports[1].device_name)
        self.assertEqual("NoneGateway-LL_HMU00_0304_flexoTHERM_PR_EBUS", reports[1].device_id)
        self.assertEqual("HEAT_PUMP", reports[1].device_type)
        self.assertEqual("COOLING", reports[1].function)
        self.assertEqual("ENVIRONMENTAL_YIELD", reports[1].energyType)
        self.assertEqual(66.0, reports[1].value)
        self.assertEqual(date(2020, 12, 8), reports[1].from_date)
        self.assertEqual(date(2020, 12, 9), reports[1].to_date)

    def test_map_hvac_no_status_messages(self) -> None:
        with open(path("files/responses/hvacstate_no_status_messages"), "r") as file:
            raw_hvacstate = json.loads(file.read())
            hvac_status = mapper.map_hvac_status(raw_hvacstate)
            self.assertEqual(2, len(hvac_status.errors))



from unittest import mock

import pytest
from aioresponses import aioresponses

from pymultimatic.api import ApiError, Connector, urls


@pytest.mark.asyncio
async def test_login_success(connector: Connector, resp: aioresponses) -> None:
    assert await connector.login()
    assert await connector.is_logged()
    assert await connector.login()


@pytest.mark.asyncio
async def test_login_auth_error(connector: Connector, raw_resp: aioresponses) -> None:
    raw_resp.post(urls.new_token(), status=200, payload={"body": {"authToken": "123"}})
    raw_resp.post(urls.authenticate(), status=401)

    try:
        await connector.login()
        raise AssertionError("Error expected")
    except ApiError as err:
        assert urls.authenticate() in err.message

    assert not await connector.is_logged()


@pytest.mark.asyncio
async def test_login_login_error(connector: Connector, raw_resp: aioresponses) -> None:
    raw_resp.post(urls.new_token(), status=401)

    try:
        await connector.login()
        assert False
    except ApiError as err:
        assert urls.new_token() in err.message

    assert not await connector.is_logged()


@pytest.mark.asyncio
async def test_auto_login_before_request(connector: Connector, resp: aioresponses) -> None:
    with mock.patch.object(connector, "login", wraps=connector.login) as mock_login:
        mock_payload = {"test": "test"}
        resp.get(urls.system(serial="123"), status=401)
        resp.get(urls.system(serial="123"), status=200, payload=mock_payload)

        assert not await connector.is_logged()
        payload = await connector.get(urls.system(serial="123"))
        assert payload == mock_payload
        assert await connector.is_logged()
        mock_login.assert_called_once()


@pytest.mark.asyncio
async def test_login_error_before_request(connector: Connector, raw_resp: aioresponses) -> None:
    with mock.patch.object(connector, "login", wraps=connector.login) as mock_login:
        raw_resp.post(urls.new_token(), status=401)
        raw_resp.get(urls.system(serial="123"), status=401)

        try:
            await connector.get(urls.system(serial="123"))
            assert False
        except ApiError as err:
            assert urls.new_token() in err.message
        mock_login.assert_called_once()
        assert not await connector.is_logged()


@pytest.mark.asyncio
async def test_delete(connector: Connector, resp: aioresponses) -> None:
    url = urls.facilities_list(serial="123")

    resp.delete(url=url, status=200)
    await connector.delete(url)


@pytest.mark.asyncio
async def test_put(connector: Connector, resp: aioresponses) -> None:
    url = urls.facilities_list(serial="123")

    resp.put(url=url, status=200)
    await connector.put(url)


@pytest.mark.asyncio
async def test_get(connector: Connector, resp: aioresponses) -> None:
    url = urls.facilities_list(serial="123")

    resp.get(url=url, status=200)
    await connector.get(url)


@pytest.mark.asyncio
async def test_post(connector: Connector, resp: aioresponses) -> None:
    url = urls.facilities_list(serial="123")

    resp.post(url=url, status=200)
    await connector.post(url)



"""Low level connector module."""
import logging
from typing import Any, Dict, Optional

import aiohttp
import attr

from . import ApiError, defaults, urls

_LOGGER = logging.getLogger(__name__)

HEADER = {
    "content-type": "application/json; charset=UTF-8",
    "Accept-Encoding": "gzip",
    "Accept": "application/json",
    "Vaillant-Mobile-App": "multiMATIC v2.1.45 b389 (Android)",
    "Cache-Control": "no-cache",
    "Pragma": "no-cache",
}


@attr.s
class Connector:
    """This is the low level smart.vaillant.com API connector.

    This is returning the raw JSON from responses or an
    :exc:`~pymultimatic.api.error.ApiError` if something goes wrong
    (basically, when response error code is 4xx or 5xx).

    On the first call, the connector will login automatically (if
    :func:`~login` was not called previously).

    On following calls, the connector will re-use the cookie received from the
    API. If the connector receives an HTTP 401, it will clear the cookie and
    try to re-login before raising an `~pymultimatic.api.error.ApiError`. This
    also means the connector is able to reconnect automatically when cookie is
    outdated.

    Please use :mod:`~pymultimatic.api.urls` in order to generate URL to be
    passed to the connector.64:36

    Args:
        user (str): User to login with.
        password (str): Password associated with the user.
        session: (aiohttp.ClientSession): Session.
        smartphone_id (str): This is required by the API to login.
    """

    _user = attr.ib(type=str)
    _password = attr.ib(type=str, repr=False)
    _session = attr.ib(type=aiohttp.ClientSession)
    _smartphone_id = attr.ib(type=str, default=defaults.SMARTPHONE_ID)

    async def login(self, force: bool = False) -> bool:
        """Log in to the API.

        By default, the ``connector`` will try to re-use cookies

        Args:
            force (bool): If set to ``True``, the connector will clear
                the cookie (if any) and start a new authentication, otherwise
                it will re-use the existing cookie.

        Returns:
            bool: True/False if authentication succeeded or not.
        """
        _LOGGER.debug(f"Will login with user {self._user}, force={force}")

        if force:
            self._clear_cookies()

        if self._get_cookies():
            _LOGGER.debug("Already logged in")
            return True

        _LOGGER.debug("Not logged in")
        token = await self._token()
        await self._authenticate(token)
        return True

    async def is_logged(self) -> bool:
        """Check if the connector is already logged in.

        It relies on the presence of cookies.
        """
        return len(self._get_cookies()) > 0

    async def logout(self) -> bool:
        """Get logged out of the API.

        It first sends a ``logout`` request to the API (it means cookies are
        invalidated), then cookies will be cleared, regardless of the result
        of the ``logout`` request.

        The connector will have to request a new token and ask for cookies if
        a new request is done.


        Raises:
            ApiError: When something went wrong with the API call.
        """
        try:
            if self._get_cookies():
                await self.post(urls.logout())
            return True
        finally:
            self._clear_cookies()

    async def _token(self) -> str:
        _LOGGER.debug("Will request token")
        params = {
            "smartphoneId": self._smartphone_id,
            "username": self._user,
            "password": self._password,
        }

        token_res = await self._session.post(url=urls.new_token(), json=params, headers=HEADER)
        if token_res.status == 200:
            json = await token_res.json()
            return str(json["body"]["authToken"])
        raise ApiError(
            f"Cannot get token at {urls.new_token()}",
            response=await token_res.text(),
            status=token_res.status,
        )

    async def _authenticate(self, token: str) -> None:
        _LOGGER.debug("Will authenticate")
        params = {
            "smartphoneId": self._smartphone_id,
            "username": self._user,
            "authToken": token,
        }

        auth_res = await self._session.post(url=urls.authenticate(), json=params, headers=HEADER)

        if auth_res.status > 399:
            raise ApiError(
                f"Unable to authenticate at {urls.authenticate()}",
                response=await auth_res.text(),
                status=auth_res.status,
            )
        _LOGGER.debug("Authentication successful")

    def _get_cookies(self) -> Dict[Any, Any]:
        return self._session.cookie_jar.filter_cookies(urls.base())  # type: ignore

    def _clear_cookies(self) -> None:
        _LOGGER.debug("Clear cookies")
        self._session.cookie_jar.clear()

    async def get(self, url: str, payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a get against vaillant API."""
        return await self.request("get", url, payload)

    async def delete(self, url: str, payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a delete against vaillant API."""
        return await self.request("delete", url, payload)

    async def put(self, url: str, payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a put against vaillant API."""
        return await self.request("put", url, payload)

    async def post(self, url: str, payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a post against vaillant API."""
        return await self.request("post", url, payload)

    async def request(self, method: str, url: str, payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a request against vaillant API."""
        _LOGGER.debug(f"Will call API: {method} {url} with payload {payload}")
        async with self._session.request(method, url, json=payload, headers=HEADER) as resp:
            if resp.status == 401:
                _LOGGER.debug(f"Request ({method}) to {url} failed, will re login")
                await self.login(True)
                return await self.request(method, url, payload)

            if resp.status > 399:
                # fetch response body, so it's available later on,
                # since this is an error, this is not always json
                await resp.read()
                raise ApiError(
                    "Cannot " + method + " " + url,
                    response=await resp.text(),
                    payload=payload,
                    status=resp.status,
                )

            data = await resp.json(content_type=None)
            _LOGGER.debug(f"Request ({method}) to {url} successful, response is: {data}")
            return data



"""Full system coming from vaillant API."""
from typing import List, Optional

import attr

from .. import utils
from . import (
    ActiveMode,
    Circulation,
    Dhw,
    FacilityDetail,
    HolidayMode,
    HotWater,
    HvacStatus,
    QuickMode,
    Report,
    Room,
    Ventilation,
    Zone,
)


@attr.s
class System:
    """This class represents the main class to manipulate vaillant system. It
    groups all the information about the system.

    Args:
        holiday (HolidayMode): Holiday mode.
        quick_mode (QuickMode):  If any quick mode is running, it's available
            here.
        zones (List[Zone]): List of zone.
        rooms (List[Room]): List of room.
        dhw (Dhw): If Dhw is available, you can find hot water and circulation.
        reports (List[Report]): sensor value coming from livereport.
        hvac_status (HvacStatus): Status of HVAC.
        facility_detail: (FacilityDetail): Detail of the facility.
        outdoor_temperature (float): Outdoor temperature
        ventilation (Ventilation): Ventilation, if any
        gateway (str): The gateway type
    """

    holiday = attr.ib(type=HolidayMode, default=HolidayMode(False))
    quick_mode = attr.ib(type=Optional[QuickMode], default=None)
    zones = attr.ib(type=List[Zone], default=[])
    rooms = attr.ib(type=List[Room], default=[])
    dhw = attr.ib(type=Optional[Dhw], default=None)
    reports = attr.ib(type=List[Report], default=[])
    hvac_status = attr.ib(type=HvacStatus, default=None)
    facility_detail = attr.ib(type=FacilityDetail, default=None)
    outdoor_temperature = attr.ib(type=Optional[float], default=None)
    ventilation = attr.ib(type=Optional[Ventilation], default=None)
    gateway = attr.ib(type=str, default=None)

    def get_active_mode_zone(self, zone: Zone) -> Optional[ActiveMode]:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.Zone`. This is the only way to
        get the real one.

        Args:
            zone (Zone): The zone you want the active mode for.

        Returns:
            ActiveMode: The active mode.
        """
        return utils.active_mode_for(zone, self.holiday, self.quick_mode)

    def get_active_mode_room(self, room: Room) -> Optional[ActiveMode]:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.Room`. This is the only way to
        get the real one.

        Args:
            room (Room): The room you want the active mode for.

        Returns:
            ActiveMode: The active mode.
        """
        return utils.active_mode_for(room, self.holiday, self.quick_mode)

    def get_active_mode_circulation(
        self, circulation: Optional[Circulation] = None
    ) -> Optional[ActiveMode]:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.Circulation`. This is the only
        way to get the real one.

        Args:
            circulation (Circulation): The circulation you want the active mode
            for. If ``None`` is passed, it will pick up the circulation from
            the system.

        Returns:
            ActiveMode: The active mode.
        """
        if not circulation and self.dhw and self.dhw.circulation:
            circulation = self.dhw.circulation

        return utils.active_mode_for(circulation, self.holiday, self.quick_mode)

    def get_active_mode_hot_water(
        self, hot_water: Optional[HotWater] = None
    ) -> Optional[ActiveMode]:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.component.HotWater`. This is the only way
        to get the real one.

        Args:
            hot_water (HotWater): The hot water you want the active mode
            for. If ``None`` is passed, it will pick up the hot water from the
            system.

        Returns:
            ActiveMode: The active mode.
        """
        if not hot_water and self.dhw and self.dhw.hotwater:
            hot_water = self.dhw.hotwater

        return utils.active_mode_for(hot_water, self.holiday, self.quick_mode)

    def get_active_mode_ventilation(self) -> Optional[ActiveMode]:
        """Get the current
        :class:`~pymultimatic.model.mode.ActiveMode` for a
        :class:`~pymultimatic.model.Ventilation`. This is the only way
        to get the real one.

        Returns:
            ActiveMode: The active mode.
        """
        return utils.active_mode_for(self.ventilation, self.holiday, self.quick_mode)



"""Tests for hvac status."""
import unittest

from pymultimatic.model import HvacStatus


class HvacStatusTest(unittest.TestCase):
    """Test class."""

    def test_online(self) -> None:
        """Test online."""
        status = HvacStatus(online="ONLINE", update="")
        self.assertTrue(status.is_online)

    def test_not_online(self) -> None:
        """Test online."""
        status = HvacStatus(online="blah", update="")
        self.assertFalse(status.is_online)

    def test_not_update(self) -> None:
        """Test online."""
        status = HvacStatus(online="", update="blah")
        self.assertFalse(status.is_online)

    def test_update(self) -> None:
        """Test online."""
        status = HvacStatus(online="", update="UPDATE_NOT_PENDING")
        self.assertTrue(status.is_up_to_date)



"""Groups everything related to a room."""
from datetime import datetime
from typing import List, Optional

import attr

from . import ActiveMode, Component, Function, OperatingModes, constants


@attr.s
class Device:
    """This is a physical device inside a :class:`Room`. It can be a VR50
    VR51 or VR52.

    Args:
        name (str): Name of the device, this is set by the mobile app.
        sgtin (str): The is the serial number of the device.
        device_type (str): Device type (VR51 = 'VALVE').
        battery_low (bool): Indicate if the device is running low battery.
        radio_out_of_reach (bool): Indicate if the device is connected to the
            system.
    """

    name = attr.ib(type=str)
    sgtin = attr.ib(type=str)
    device_type = attr.ib(type=str)
    battery_low = attr.ib(type=bool)
    radio_out_of_reach = attr.ib(type=bool)


@attr.s
class Room(Function, Component):
    """This is representing a room from the system.

    Note:
        The boost function - *the boost function is activated by pressing the
        temperature selector. This means that the heating valves on the
        assigned radiator thermostats are immediately opened 80% for 5
        minutes* - is not reflected into the API.

    Args:
        humidity (float): In case of a VR51 is inside the room, you can get
            humidity through the API.
        child_lock (bool): Indicate if the device is locked, meaning you cannot
            use the buttons on the device.
        window_open (bool): Indicate if a window is open in the room. Of
            course, this is data interpretation from vaillant API.
        devices (List[Device]): List of :class:`Device` inside the room.
    """

    MODES = [
        OperatingModes.OFF,
        OperatingModes.MANUAL,
        OperatingModes.AUTO,
        OperatingModes.QUICK_VETO,
    ]
    """List of mode that are applicable to rooms component."""

    MIN_TARGET_TEMP = constants.FROST_PROTECTION_TEMP
    """Min `target temperature` that can be apply to a room."""

    MAX_TARGET_TEMP = constants.THERMOSTAT_MAX_TEMP
    """Max `target temperature` that can be apply to a room."""

    target_low = attr.ib(default=None, init=False)
    humidity = attr.ib(type=Optional[float], default=None)
    child_lock = attr.ib(type=bool, default=None)
    window_open = attr.ib(type=bool, default=None)
    devices = attr.ib(type=List[Device], default=None)

    @property
    def active_mode(self) -> ActiveMode:
        """ActiveMode: Get the :class:`~pymultimatic.model.mode.ActiveMode`
        for this function. All operating modes are handled,
        **but not quick veto nor quick mode.**
        """
        if self.quick_veto:
            mode = ActiveMode(self.quick_veto.target, OperatingModes.QUICK_VETO)

        elif self.operating_mode == OperatingModes.AUTO:
            setting = self.time_program.get_for(datetime.now())
            mode = ActiveMode(setting.target_temperature, OperatingModes.AUTO, setting.setting)

        elif self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(self.MIN_TARGET_TEMP, OperatingModes.OFF)
        else:  # MODE_MANUAL
            mode = ActiveMode(self.target_high, OperatingModes.MANUAL)

        return mode

    def _active_mode(self) -> ActiveMode:
        pass



"""Mapped model from the API."""
from .mode import (  # noqa: F401
    Mode,
    OperatingMode,
    OperatingModes,
    QuickVeto,
    ActiveMode,
    SettingMode,
    SettingModes,
)
from .timeprogram import TimeProgram, TimeProgramDay, TimePeriodSetting  # noqa: F401
from .common import Component, Function  # noqa: F401
from .zone import ZoneCooling, ZoneHeating, Zone, ActiveFunction  # noqa: F401
from .room import Device, Room  # noqa: F401
from .status import HvacStatus, BoilerStatus, Error  # noqa: F401
from .syncstate import SyncState  # noqa: F401
from .dhw import Dhw, HotWater, Circulation  # noqa: F401
from .report import Report, EmfReport  # noqa: F401
from .ventilation import Ventilation  # noqa: F401
from .quick_mode import QuickMode, QuickModes, HolidayMode  # noqa: F401
from .info import FacilityDetail  # noqa: F401
from .system import System  # noqa: F401



"""Mappers from json to model classes."""
from datetime import datetime, timedelta
from typing import Any, List, Optional, Tuple

from . import (
    ActiveFunction,
    BoilerStatus,
    Circulation,
    Device,
    Dhw,
    Error,
    FacilityDetail,
    HolidayMode,
    HotWater,
    HvacStatus,
    OperatingMode,
    OperatingModes,
    QuickMode,
    QuickModes,
    QuickVeto,
    Report,
    Room,
    SettingModes,
    SyncState,
    TimePeriodSetting,
    TimeProgram,
    TimeProgramDay,
    Ventilation,
    Zone,
    ZoneCooling,
    ZoneHeating,
    EmfReport,
)

_DATE_FORMAT = "%Y-%m-%d"
_DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
_DAYS_OF_WEEK = [
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
]


def map_emf_reports(json) -> List[EmfReport]:
    """Map emf reports"""
    reports = []
    if json and json.get("body"):
        for raw_devices in json.get("body"):
            device_id = raw_devices.get("id")
            device_type = raw_devices.get("type")
            device_name = raw_devices.get("marketingName")
            for raw_report in raw_devices.get("reports"):
                reports.append(
                    EmfReport(
                        device_id,
                        device_name,
                        device_type,
                        raw_report.get("function"),
                        raw_report.get("energyType"),
                        raw_report.get("currentMeterReading"),
                        datetime.strptime(raw_report.get("from"), _DATE_FORMAT).date(),
                        datetime.strptime(raw_report.get("to"), _DATE_FORMAT).date(),
                    )
                )
    return reports


def map_gateway(json) -> str:
    """Map gateway."""
    return str(json.get("body", {}).get("gatewayType", None))


def map_quick_mode_from_system(full_system) -> Optional[QuickMode]:
    """Map *quick mode*."""
    if full_system:
        quick_mode = full_system.get("body", {}).get("configuration", {}).get("quickmode")
        return _map_quick_mode(quick_mode) if quick_mode else None
    return None


def map_quick_mode(json) -> Optional[QuickMode]:
    """Map quick mode."""
    return _map_quick_mode(json.get("body"))


def _map_quick_mode(json) -> Optional[QuickMode]:
    if json.get("quickmode"):
        mode = QuickModes.get(json.get("quickmode"), json.get("duration"))
        return mode if mode != QuickModes.QUICK_VETO else None
    return None


def map_holiday_mode(json) -> HolidayMode:
    """Map *holiday mode*."""
    return _map_holiday_mode(json.get("body"))


def map_holiday_mode_from_system(json) -> HolidayMode:
    """Map *holiday mode*."""
    return _map_holiday_mode(json.get("body", {}).get("configuration", {}).get("holidaymode"))


def _map_holiday_mode(json) -> HolidayMode:
    """Map *holiday mode*."""
    mode = HolidayMode(False)
    if json:
        mode.is_active = bool(json.get("active"))
        target = json.get("temperature_setpoint")
        mode.target = float(target) if target else None
        start_date = json.get("start_date")
        end_date = json.get("end_date")
        mode.start_date = datetime.strptime(start_date, _DATE_FORMAT).date() if start_date else None
        mode.end_date = datetime.strptime(end_date, _DATE_FORMAT).date() if end_date else None
    return mode


def map_hvac_status(json) -> HvacStatus:
    """Map hvac status."""
    meta = json.get("meta", {})
    return HvacStatus(
        online=meta.get("onlineStatus", {}).get("status"),
        update=meta.get("firmwareUpdateStatus", {}).get("status"),
        boiler_status=_map_boiler_status(json),
        errors=map_errors(json),
    )


def map_outdoor_temp(json) -> Optional[float]:
    """get *outdoor_temperature*."""
    tmp = json.get("body", {}).get("outside_temperature", None)
    return float(tmp) if tmp else None


def map_outdoor_temp_from_system(json) -> Optional[float]:
    """get *outdoor_temperature*."""
    tmp = json.get("body", {}).get("status", {}).get("outside_temperature", None)
    return float(tmp) if tmp else None


def map_rooms(raw_rooms) -> List[Room]:
    """Map *rooms*."""

    rooms: List[Room] = []
    if raw_rooms:
        rooms = [map_room(raw_room) for raw_room in raw_rooms.get("body", {}).get("rooms", [])]
    return rooms


def map_room(raw_room) -> Optional[Room]:
    """Map *room*."""
    if raw_room:
        raw_room = raw_room.get("body", raw_room)

        if raw_room:
            config = raw_room.get("configuration", {})

            t_prog, mode, high, low = _map_function(raw_room)

            room_id = raw_room.get("roomIndex")
            child_lock = config.get("childLock")
            current_temp = config.get("currentTemperature")
            devices = map_devices(config.get("devices"))
            window_open = config.get("isWindowOpen")
            name = config.get("name")
            humidity = config.get("currentHumidity")

            raw_quick_veto = config.get("quickVeto")
            quick_veto = None
            if raw_quick_veto:
                quick_veto = QuickVeto(
                    raw_quick_veto.get("remainingDuration"),
                    config.get("temperatureSetpoint"),
                )

            return Room(
                id=room_id,
                name=name,
                time_program=t_prog,
                temperature=current_temp,
                target_high=high,
                operating_mode=mode,
                quick_veto=quick_veto,
                child_lock=child_lock,
                window_open=window_open,
                devices=devices,
                humidity=humidity,
            )
    return None


def map_devices(raw_devices) -> List[Device]:
    """Map *devices* of a room."""
    devices = []
    if raw_devices:
        for raw_device in raw_devices:
            name = raw_device.get("name")
            device_type = raw_device.get("deviceType")
            battery_low = raw_device.get("isBatteryLow")
            radio_out_of_reach = raw_device.get("isRadioOutOfReach")
            sgtin = raw_device.get("sgtin")
            devices.append(Device(name, sgtin, device_type, battery_low, radio_out_of_reach))

    return devices


def map_time_program(raw_time_program, key: Optional[str] = None) -> TimeProgram:
    """Map *time program*."""
    result = {}
    if raw_time_program:
        for day_of_week in _DAYS_OF_WEEK:
            result[day_of_week] = map_time_program_day(raw_time_program.get(day_of_week), key)

    return TimeProgram(result)


def map_time_program_day(raw_time_program_day, key: Optional[str] = None) -> TimeProgramDay:
    """Map *time program day* and *time program day settings*."""
    settings = []
    if raw_time_program_day:
        # By default in multimatic mode
        multimatic_timeprogram = True
        for time_setting in raw_time_program_day:
            # Try if multimatic or senso
            multimatic_timeprogram = "startTime" in time_setting
            mode = None
            end_time = None
            if multimatic_timeprogram:
                start_time = time_setting.get("startTime")
                target_temp = time_setting.get("temperatureSetpoint")
                if key:
                    mode = SettingModes.get(time_setting.get(key))
            else:
                start_time = time_setting.get("start_time")
                end_time = time_setting.get("end_time")
                target_temp = time_setting.get("setpoint")
                mode = SettingModes.DAY

            settings.append(TimePeriodSetting(start_time, target_temp, mode, end_time))

        time_program_day = TimeProgramDay(settings)
        # In senso mode, deactivated periods are not declared.
        if not multimatic_timeprogram:
            time_program_day.complete_empty_periods(SettingModes.NIGHT)

    return time_program_day


def _map_boiler_status(hvac_state) -> Optional[BoilerStatus]:
    """Map *boiler status."""
    if hvac_state:
        hvac_state_info = _find_hvac_message_status(hvac_state)
        if hvac_state_info:
            last_update = _datetime_mandatory(hvac_state_info.get("timestamp"))
            device_name = str(hvac_state_info.get("deviceName"))
            code = str(hvac_state_info.get("statusCode"))
            title = str(hvac_state_info.get("title"))
            description = str(hvac_state_info.get("description"))
            hint = str(hvac_state_info.get("hint"))
            return BoilerStatus(device_name, title, code, description, last_update, hint)

    return None


def map_facility_detail(facilities, serial=None) -> FacilityDetail:
    """Map facility detail."""
    facilities_list = facilities.get("body", {}).get("facilitiesList", [])

    if serial:
        facility = [
            facility for facility in facilities_list if facility.get("serialNumber") == serial
        ][0]
    else:
        facility = facilities_list[0]
    return FacilityDetail(
        name=facility.get("name", None),
        serial_number=facility.get("serialNumber"),
        firmware_version=facility.get("firmwareVersion", None),
        ethernet_mac=facility.get("networkInformation", {}).get("macAddressEthernet"),
        wifi_mac=facility.get("networkInformation", {}).get("macAddressWifiAccessPoint"),
    )


def map_zones(json) -> List[Zone]:
    """Map *zones*."""
    return _map_zones(json.get("body", []))


def map_zones_from_system(json) -> List[Zone]:
    """Map *zones*."""
    return _map_zones(json.get("body", {}).get("zones", []))


def _map_zones(json) -> List[Zone]:
    """Map *zones*."""
    zones = []
    for raw_zone in json:
        zone = map_zone(raw_zone)
        if zone:
            zones.append(zone)
    return zones


def map_zone(raw_zone) -> Optional[Zone]:
    """Map *zones*."""
    raw_zone_body = raw_zone.get("body")

    if raw_zone_body:
        raw_zone = raw_zone_body

    if raw_zone:
        zone_id = raw_zone.get("_id")
        configuration = raw_zone.get("configuration", {})
        name = configuration.get("name", "").strip()
        temperature = configuration.get("inside_temperature")
        active_function = ActiveFunction[
            configuration.get("active_function", ActiveFunction.STANDBY.name)
        ]
        quick_veto = _map_quick_veto_zone(configuration.get("quick_veto"))

        currently_controlled_by = raw_zone.get("currently_controlled_by")
        rbr = currently_controlled_by and (
            (isinstance(currently_controlled_by, str) and currently_controlled_by == "RBR")
            or (
                isinstance(currently_controlled_by, dict)
                and currently_controlled_by.get("name", "") == "RBR"
            )
        )

        raw_heating = raw_zone.get("heating", {})
        raw_cooling = raw_zone.get("cooling", {})
        enabled = configuration.get("enabled", bool())

        zone_cooling = None
        func = _map_function(raw_heating, "setting", rbr, quick_veto is not None)
        zone_heating = ZoneHeating(func[0], func[1], func[2], func[3])

        if raw_cooling:
            func = _map_function(raw_cooling, "setting")
            zone_cooling = ZoneCooling(func[0], func[1], func[2], func[3])

        return Zone(
            id=zone_id,
            name=name,
            temperature=temperature,
            quick_veto=quick_veto,
            active_function=active_function,
            rbr=rbr,
            heating=zone_heating,
            cooling=zone_cooling,
            enabled=enabled,
        )
    return None


def map_ventilation_from_system(system) -> Optional[Ventilation]:
    """Maps *ventilation*."""
    ventilation = None
    if system:
        fans = system.get("body", {}).get("ventilation", [])
        if fans:
            ventilation = _map_ventilation(fans[0])
    return ventilation


def map_ventilation(json) -> Optional[Ventilation]:
    """Maps *ventilation*."""
    return _map_ventilation(json.get("body")[0])


def _map_ventilation(json) -> Optional[Ventilation]:
    tprog, mode, target_high, target_low = _map_function(json.get("fan", {}), "setting")
    fan_id = json.get("_id")
    return Ventilation(
        id=fan_id,
        name="Ventilation",
        time_program=tprog,
        operating_mode=mode,
        target_high=target_high,
        target_low=target_low,
    )


def _map_function(
    raw, tp_key=None, rbr=False, quick_veto=False
) -> Tuple[TimeProgram, OperatingMode, float, float]:
    conf = raw.get("configuration", {})
    mode = conf.get("mode")
    if not mode:
        mode = conf.get("operation_mode")
        if not mode:
            mode = conf.get("operationMode")

    operating_mode: OperatingMode
    if rbr:
        operating_mode = OperatingModes.get(mode) if mode else None
    elif quick_veto:
        operating_mode = OperatingModes.QUICK_VETO
    else:
        operating_mode = OperatingModes.get(mode)

    target_high = conf.get("setpoint_temperature", None)
    if not target_high:
        target_high = conf.get("temperature_setpoint", None)
        if not target_high:
            target_high = conf.get("temperatureSetpoint", None)
        if not target_high:
            target_high = conf.get("day_level", None)
        if not target_high:
            target_high = conf.get("manual_mode_temperature_setpoint", None)
        if not target_high:
            target_high = conf.get("hotwater_temperature_setpoint", None)

    target_low = conf.get("setback_temperature", None)
    if not target_low:
        target_low = conf.get("night_level", None)
    if not target_low:
        target_low = conf.get("setback_temperature_setpoint", None)

    time_program = map_time_program(raw.get("timeprogram"), tp_key)

    return time_program, operating_mode, target_high, target_low


def map_hot_water_from_system(full_system, live_reports) -> Optional[HotWater]:
    """Map *hot water*."""
    dhws = full_system.get("body", {}).get("dhw")

    if dhws:
        # In Senso API, dhws is not a array
        if isinstance(dhws, dict):
            dhws = [dhws]
        hotwater = dhws[0].get("hotwater")
        # Senso API does not provide an ID
        dhw_id = dhws[0].get("_id", "Control_DHW")
        temp = _get_report_value(_find_dhw_temperature_report(live_reports))
        return _map_hot_water(hotwater, dhw_id, temp)

    return None


def map_hot_water_from_dhw(json) -> Optional[HotWater]:
    """Mapp hotware from dhw."""
    dhws = json.get("body", [])
    # In Senso API, dhws is not a array
    if isinstance(dhws, dict):
        dhws = [dhws]
    dhw = dhws[0]
    # Senso API does not provide an ID
    dhw_id = dhw.get("_id", "Control_DHW")
    return _map_hot_water(dhw.get("hotwater"), dhw_id, None)


def map_hot_water(json, dhw_id: str) -> Optional[HotWater]:
    """Map *hot water*."""
    if json:
        hotwater = json.get("body", {})
        return _map_hot_water(hotwater, dhw_id, None)
    return None


def map_dhw_from_system(full_system, live_reports) -> Dhw:
    """Map *dhw*."""
    circulation = map_circulation_from_system(full_system)
    hotwater = map_hot_water_from_system(full_system, live_reports)
    return Dhw(hotwater=hotwater, circulation=circulation)


def map_dhw(dhw) -> Dhw:
    """Map *dhw*."""
    circulation = map_circulation_from_dhw(dhw)
    hotwater = map_hot_water_from_dhw(dhw)
    return Dhw(hotwater=hotwater, circulation=circulation)


def map_circulation_from_dhw(json) -> Optional[Circulation]:
    """Map *circulation*."""
    if json:
        dhws = json.get("body", [])
        # In Senso API, dhws is not a array
        if isinstance(dhws, dict):
            dhws = [dhws]

        if dhws:
            circulation = dhws[0].get("circulation")
            # Senso API does not provide an ID
            dhw_id = dhws[0].get("_id", "Control_DHW")
            return _map_circulation(circulation, dhw_id)
    return None


def map_circulation_from_system(full_system) -> Optional[Circulation]:
    """Map *circulation*."""
    if full_system:
        hot_water_list = full_system.get("body", {}).get("dhw", [])
        # In Senso API, hot water is not a array
        if isinstance(hot_water_list, dict):
            hot_water_list = [hot_water_list]

        if hot_water_list:
            raw_circulation = hot_water_list[0].get("circulation")
            # Senso API does not provide an ID
            dhw_id = hot_water_list[0].get("_id", "Control_DHW")

            if raw_circulation:
                return _map_circulation(raw_circulation, dhw_id)
    return None


def map_circulation_alone(raw_circulation, dhw_id: str) -> Optional[Circulation]:
    """Map *circulation*."""
    if raw_circulation:
        raw_circulation_body = raw_circulation.get("body", {})
        return _map_circulation(raw_circulation_body, dhw_id)
    return None


def map_errors(hvac_state) -> List[Error]:
    """Map *errors*."""
    errors = []
    for error in hvac_state.get("body", {}).get("errorMessages", []):
        if error.get("type") != "STATUS":
            errors.append(
                Error(
                    error.get("deviceName"),
                    error.get("title"),
                    error.get("statusCode"),
                    error.get("description"),
                    _datetime_mandatory(error.get("timestamp")),
                )
            )
    return errors


def map_hvac_sync_state(hvac_state) -> Optional[SyncState]:
    """Map sync state."""
    if hvac_state:
        states = hvac_state.get("meta", {}).get("syncState", [])
        if states:
            return _map_state(states[0])
    return None


def map_serial_number(facilities) -> str:
    """Map serial number."""
    facility = facilities.get("body", {}).get("facilitiesList", [])[0]
    return str(facility.get("serialNumber", None))


def map_systemcontrol(facilities) -> str:
    """Map system control found in capabilities."""
    facility = facilities.get("body", {}).get("facilitiesList", [])[0]
    capabilities = facility.get("capabilities", [])
    return str([s for s in capabilities if s.startswith("SYSTEMCONTROL_")][0])


def _map_state(raw_state) -> Optional[SyncState]:
    state = str(raw_state.get("state"))
    timestamp = _datetime_mandatory(raw_state.get("timestamp"))
    link = raw_state.get("link", {}).get("resourceLink")
    return SyncState(state, timestamp, link)


def map_reports(live_report) -> List[Report]:
    """Maps *Reports*."""
    reports = []

    if live_report:
        for device in live_report.get("body", {}).get("devices", []):
            device_id = device.get("_id")
            device_name = device.get("name")

            for report_json in device.get("reports", []):
                report = _map_report(report_json)
                if report:
                    report.device_id = device_id
                    report.device_name = device_name
                    reports.append(report)

    return reports


def map_report(json) -> Optional[Report]:
    """Map report."""
    return _map_report(json.get("body"))


def _map_report(json) -> Optional[Report]:
    if json:
        return Report(
            id=json.get("_id"),
            value=json.get("value"),
            name=json.get("name"),
            unit=json.get("unit"),
        )
    return None


def _map_hot_water(raw_hot_water, dhw_id, current_temp) -> Optional[HotWater]:
    timep, mode, high, low = _map_function(raw_hot_water, "mode")
    return HotWater(
        id=dhw_id,
        name="hotwater",
        time_program=timep,
        temperature=current_temp,
        target_high=high,
        operating_mode=mode,
    )


def _map_circulation(raw_circulation, dhw_id: str) -> Circulation:
    func = _map_function(raw_circulation, "setting")

    return Circulation(id=dhw_id, name="Circulation", time_program=func[0], operating_mode=func[1])


def _find_hvac_message_status(hvac_state) -> Optional[Any]:
    for message in hvac_state.get("body", {}).get("errorMessages", []):
        if message.get("type") == "STATUS":
            return message
    return None


def _find_dhw_temperature_report(live_report) -> Optional[Any]:
    for device in live_report.get("body", {}).get("devices", []):
        for report in device.get("reports", []):
            if (
                report.get("associated_device_function") == "DHW"
                and report.get("_id") == "DomesticHotWaterTankTemperature"
            ):
                return report
    return None


def _get_report_value(report) -> Any:
    if report:
        body = report.get("body")
        if body:
            return body.get("value")
        return report.get("value")
    return None


def _map_quick_veto_zone(raw_quick_veto) -> Optional[QuickVeto]:
    if raw_quick_veto and ("active" not in raw_quick_veto or raw_quick_veto.get("active")):
        # No way to find the start date, the quick zone veto lasts 6 hours for Multimatic.
        # For Senso, the expiry date is not immediately present.
        expires_at = raw_quick_veto.get("expires_at")
        duration = None
        if expires_at:
            expires_at_datetime = datetime.strptime(expires_at, _DATE_TIME_FORMAT)
            duration, _ = divmod(expires_at_datetime - datetime.utcnow(), timedelta(minutes=1))
            if duration <= 0:
                return None
        target = raw_quick_veto.get("setpoint_temperature")
        if not target:
            target = raw_quick_veto.get("temperature_setpoint")
        return QuickVeto(target=target, duration=duration)
    return None


def _datetime(timestamp: Optional[int]) -> Optional[datetime]:
    if timestamp:
        return datetime.fromtimestamp(timestamp / 1000)
    return None


def _datetime_mandatory(timestamp: int) -> datetime:
    return datetime.fromtimestamp(timestamp / 1000)



import unittest
from datetime import datetime

from pymultimatic.model import SyncState


class StateTest(unittest.TestCase):
    def test_is_sync(self) -> None:
        self.assertTrue(SyncState("SYNCED", datetime.now(), "link").is_synced)

    def test_is_pending(self) -> None:
        self.assertTrue(SyncState("PENDING", datetime.now(), "link").is_pending)

    def test_is_outdated(self) -> None:
        self.assertTrue(SyncState("OUTDATED", datetime.now(), "link").is_outdated)

    def test_is_init(self) -> None:
        self.assertTrue(SyncState("INITIALIZING", datetime.now(), "link").is_init)



"""Convenient manager to easily gets data from API."""
import asyncio
import logging
from datetime import date, timedelta, datetime
from typing import Any, Callable, List, Optional, Tuple, Type

from aiohttp import ClientSession
from schema import Schema, SchemaError

from .api import ApiError, Connector, WrongResponseError, defaults, schemas
from .model import (
    Circulation,
    Dhw,
    FacilityDetail,
    HolidayMode,
    HotWater,
    HvacStatus,
    OperatingMode,
    OperatingModes,
    QuickMode,
    QuickVeto,
    Report,
    Room,
    System,
    Ventilation,
    Zone,
    ZoneCooling,
    ZoneHeating,
    constants,
    mapper,
    EmfReport,
)

_LOGGER = logging.getLogger(__name__)


def ignore_http_409(return_value: Any = None) -> Callable[..., Any]:
    """Ignore ApiError if status code is 409."""

    def decorator(func: Callable[..., Any]) -> Any:
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return await func(*args, **kwargs)
            except ApiError as ex:
                if ex.status != 409:
                    raise
                return return_value

        return wrapper

    return decorator


def retry_async(
    num_tries: int = 5,
    on_exceptions: Tuple[Type[BaseException], ...] = (Exception,),
    on_status_codes: Tuple[int, ...] = (),
    backoff_base: float = 0.5,
) -> Callable[..., Any]:
    """In case of exceptions, retries decorated async function multiple times.

    Uses increasing backoff between tries.

    Args:
         num_tries (int): Max number of tries.
         on_exceptions (tuple): Retries on specific exceptions only.
         on_status_codes (tuple): If `ApiError` occurs,
            retry only on specified status codes.
         backoff_base (float): Backoff base value.
    """
    on_exceptions = on_exceptions + (ApiError,)

    def decorator(func: Callable[..., Any]) -> Any:
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            _num_tries = num_tries
            last_response: Optional[str] = None
            while _num_tries > 0:
                _num_tries -= 1
                try:
                    return await func(*args, **kwargs)
                except on_exceptions as ex:
                    if not _num_tries:
                        if isinstance(ex, ApiError):
                            raise ex
                        raise ApiError(
                            "Cannot get correct response",
                            response=last_response,
                            status=200,
                        ) from ex
                    if isinstance(ex, ApiError):
                        last_response = ex.response
                        if ex.status not in on_status_codes:
                            raise
                    retry_in = backoff_base * (num_tries - _num_tries)
                    _LOGGER.debug(f"Error occurred, retrying in {retry_in}", exc_info=True)
                    await asyncio.sleep(retry_in)

        return wrapper

    return decorator


class SystemManager:
    """This is a convenient manager to help interact with vaillant API.

    The manager is calling the
    :class:`~pymultimatic.api.connector.ApiConnector`, the manager will
    provide to correct :mod:`~pymultimatic.api.payloads` and
    :mod:`~pymultimatic.api.urls` according to what you want to do.

    The manager is throwing :class:`~pymultimatic.api.error.ApiError`
    without altering it.

    Args:
        user (str): User to login with.
        password (str): Password associated with the user.
        session: (aiohttp.ClientSession): Session.
        smartphone_id (str): This is required by the API to login.
        serial (str): If you have multiple facilities,
            you can specify which one to access
        application (str): Multimatic or senso app (default multimatic)
    """

    def __init__(
        self,
        user: str,
        password: str,
        session: ClientSession,
        smartphone_id: str = defaults.SMARTPHONE_ID,
        serial: Optional[str] = None,
        application: Optional[str] = defaults.MULTIMATIC,
    ):
        self._connector: Connector = Connector(user, password, session, smartphone_id)
        self._serial = serial
        self._fixed_serial = self._serial is not None
        self._ensure_ready_lock = asyncio.Lock()
        self._application = defaults.SENSO if application == defaults.SENSO else defaults.MULTIMATIC

        if application == defaults.SENSO:
            self.urls = __import__("pymultimatic.api.urls_senso", fromlist=[""])
            self.payloads = __import__("pymultimatic.api.payloads_senso", fromlist=[""])
        else:
            self.urls = __import__("pymultimatic.api.urls", fromlist=[""])
            self.payloads = __import__("pymultimatic.api.payloads", fromlist=[""])

    async def login(self, force_login: bool = False) -> bool:
        """Try to login to the API, see
        :func:`~pymultimatic.api.connector.ApiConnector.login`

        Args:
            force_login (bool): Whether the login should be forced or not.

        Returns:
            bool: True/False if authentication succeeded or not.
        """
        return await self._connector.login(force_login)

    async def logout(self) -> None:
        """Get logged out from the API, see
        :func:`~pymultimatic.api.connector.ApiConnector.logout`
        """
        if not self._fixed_serial:
            self._serial = None
        await self._connector.logout()

    async def get_system(self) -> System:
        """Get the full :class:`~pymultimatic.model.system.System`. It may
        take some times, it actually does multiples API calls, depending on
        your system configuration.

        Returns:
            System: the full system.
        """

        (
            facilities,
            full_system,
            live_report,
            hvac_state,
            gateway_json,
        ) = await asyncio.gather(
            self._call_api(self.urls.facilities_list, schema=schemas.FACILITIES),
            self._call_api(self.urls.system, schema=schemas.SYSTEM),
            self._call_api(self.urls.live_report, schema=schemas.LIVE_REPORTS),
            self._call_api(self.urls.hvac, schema=schemas.HVAC),
            self._call_api(self.urls.gateway_type, schema=schemas.GATEWAY),
        )

        hvac_status = mapper.map_hvac_status(hvac_state)
        holiday = mapper.map_holiday_mode_from_system(full_system)
        zones = mapper.map_zones_from_system(full_system)
        outdoor_temp = mapper.map_outdoor_temp_from_system(full_system)
        quick_mode = mapper.map_quick_mode_from_system(full_system)
        ventilation = mapper.map_ventilation_from_system(full_system)
        dhw = mapper.map_dhw_from_system(full_system, live_report)
        reports = mapper.map_reports(live_report)
        facility_detail = mapper.map_facility_detail(facilities, self._serial)
        gateway = mapper.map_gateway(gateway_json)

        rooms: List[Room] = []
        if [z for z in zones if z.rbr]:
            rooms_raw = await self._call_api(self.urls.rooms, schema=schemas.ROOM_LIST)
            rooms = mapper.map_rooms(rooms_raw)

        return System(
            holiday=holiday,
            quick_mode=quick_mode,
            zones=zones,
            rooms=rooms,
            dhw=dhw,
            reports=reports,
            outdoor_temperature=outdoor_temp,
            hvac_status=hvac_status,
            facility_detail=facility_detail,
            ventilation=ventilation,
            gateway=gateway,
        )

    @ignore_http_409(return_value=[])
    async def get_emf_devices(self) -> List[EmfReport]:
        """Get all the EMF reports available.

        Returns:
            The list of report emf devices
        """
        return mapper.map_emf_reports(await self._call_api(self.urls.emf_devices))

    async def get_gateway(self) -> str:
        """Get the gateway type (VR900, VR920, etc)

        Returns:
            The gateway type
        """
        return mapper.map_gateway(await self._call_api(self.urls.gateway_type))

    @ignore_http_409()
    async def get_outdoor_temperature(self) -> Optional[float]:
        """Get the outdoor temperature

        Returns:
            The outdoor temperature if available
        """
        return mapper.map_outdoor_temp(await self._call_api(self.urls.system_status))

    @ignore_http_409()
    async def get_hvac_status(self) -> HvacStatus:
        """Get the :class:`~pymultimatic.model.HvacStatus`

        Returns:
            The hvac status
        """
        return mapper.map_hvac_status(await self._call_api(self.urls.hvac))

    async def get_facility_detail(self, serial: Optional[str] = None) -> FacilityDetail:
        """Get the :class:`~pymultimatic.model.FacilityDetail` for a given serial

        Returns:
            The facility detail
        """
        serial = serial if serial is not None else self._serial
        return mapper.map_facility_detail(
            await self._call_api(self.urls.facilities_list, schema=schemas.FACILITIES),
            serial,
        )

    @ignore_http_409()
    async def get_live_reports(self) -> List[Report]:
        """Get available list of :class:`~pymultimatic.model.Report`

        Returns:
            A list of live reports
        """
        return mapper.map_reports(
            await self._call_api(self.urls.live_report, schema=schemas.LIVE_REPORTS)
        )

    @ignore_http_409()
    async def get_live_report(self, report_id: str, device_id: str) -> Optional[Report]:
        """Get available list of :class:`~pymultimatic.model.Report`

        Returns:
            A list of live reports
        """
        json = await self._call_api(
            self.urls.live_report_device,
            params={"device_id": device_id, "report_id": report_id},
            schema=schemas.LIVE_REPORT,
        )
        return mapper.map_report(json)

    @ignore_http_409()
    async def get_ventilation(self) -> Optional[Ventilation]:
        """Get the :class:`~pymultimatic.model.component.Ventilation`

        Returns:
            Ventilation: the ventilation
        """
        return mapper.map_ventilation(
            await self._call_api(self.urls.system_ventilation, schema=schemas.VENTILATION_LIST)
        )

    @ignore_http_409()
    async def get_holiday_mode(self) -> HolidayMode:
        """Get the :class:`~pymultimatic.model.quick_mode.HolidayMode`

        Returns:
            HolidayMode: the holiday mode
        """
        raw = await self._call_api(self.urls.system_holiday_mode)
        return mapper.map_holiday_mode(raw)

    @ignore_http_409()
    async def get_quick_mode(self) -> Optional[QuickMode]:
        """Get the :class:`~pymultimatic.model.quick_mode.QuickMode`

        Returns:
            QuickMode: the quick mode or None
        """
        return mapper.map_quick_mode(await self._call_api(self.urls.system_quickmode))

    @ignore_http_409()
    async def get_hot_water(self, dhw_id: str) -> Optional[HotWater]:
        """Get the :class:`~pymultimatic.model.component.HotWater`
        information for the given id.

        Args:
            dhw_id (str): domestic hot water id (same as
                :func:`get_circulation`).

        Returns:
            HotWater: the hot water information, if any.
        """
        dhw = await self._call_api(
            self.urls.hot_water, params={"id": dhw_id}, schema=schemas.FUNCTION
        )
        return mapper.map_hot_water(dhw, dhw_id)

    @ignore_http_409()
    async def get_dhw(self) -> Optional[Dhw]:
        """Get the :class:`~pymultimatic.model.Dhw` (Domestic Hot Water)

        Returns:
            The domestic Hot water (circulation + hot water), if any
        """
        dhw = await self._call_api(self.urls.dhws, schema=schemas.DHWS)
        return mapper.map_dhw(dhw)

    @ignore_http_409(return_value=[])
    async def get_rooms(self) -> Optional[List[Room]]:
        """Get a list of :class:`~pymultimatic.model.component.Room`

        Returns:
            Rooms: list of room
        """
        rooms = await self._call_api(self.urls.rooms, schema=schemas.ROOM_LIST)
        return mapper.map_rooms(rooms)

    @ignore_http_409()
    async def get_room(self, room_id: str) -> Optional[Room]:
        """Get the :class:`~pymultimatic.model.component.Room` information
        for the given id.

        Args:
            room_id (str): Id of the room, this is actually an index, 1,2,3,4
            ... depending of the number of rooms you have.

        Returns:
            Room: the room information, if any.
        """
        new_room = await self._call_api(self.urls.room, params={"id": room_id}, schema=schemas.ROOM)
        return mapper.map_room(new_room)

    @ignore_http_409(return_value=[])
    async def get_zones(self) -> Optional[List[Zone]]:
        """Get a list of :class:`~pymultimatic.model.component.Zone`

        Returns:
            Zones: list of Zone
        """
        rooms = await self._call_api(self.urls.zones, schema=schemas.ZONE_LIST)
        return mapper.map_zones(rooms)

    @ignore_http_409()
    async def get_zone(self, zone_id: str) -> Optional[Zone]:
        """Get the :class:`~pymultimatic.model.component.Zone` information
        for the given id.

        Args:
            zone_id (str): Name of the room, basically, you set it through the
                VRC 700.

        Returns:
            Zone: the zone information, if any.
        """
        new_zone = await self._call_api(self.urls.zone, params={"id": zone_id}, schema=schemas.ZONE)
        return mapper.map_zone(new_zone)

    @ignore_http_409()
    async def get_circulation(self, dhw_id: str) -> Optional[Circulation]:
        """Get the :class:`~pymultimatic.model.component.Circulation`
        information for the given id.

        Args:
            dhw_id (str): domestic hot water id (same as :func:`get_hot_water`)

        Returns:
            Circulation: the circulation information, if any.
        """
        new_circulation = await self._call_api(
            self.urls.circulation, params={"id": dhw_id}, schema=schemas.FUNCTION
        )
        return mapper.map_circulation_alone(new_circulation, dhw_id)

    async def set_quick_mode(self, quick_mode: QuickMode) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickMode` system wise.

        This will override the current
        :class:`~pymultimatic.model.mode.QuickMode`, if any.

        Only compatible with MULTIMATIC.
        Args:
            quick_mode (QuickMode): the quick mode to set, see
                :class:`~pymultimatic.model.mode.QuickModes`
        """
        if self._application != defaults.MULTIMATIC:
            _LOGGER.warning(f"quick_mode is not compatible with {self._application}")
        await self._call_api(
            self.urls.system_quickmode,
            payload=self.payloads.quickmode(quick_mode.name, quick_mode.duration),
        )

    @ignore_http_409(return_value=False)
    async def remove_quick_mode(self) -> bool:
        """Removes current :class:`~pymultimatic.model.mode.QuickMode`.
        Only compatible with MULTIMATIC.

        Note:
            if there is not :class:`~pymultimatic.model.mode.QuickMode` set,
            the API returns an error (HTTP 409). **This error is swallowed by
            the manager**, so you don't have to handle it.

        Returns:
              True/False: if quick_mode has been removed or not
                (if there wasn't any quick mode set previously)
        """

        if self._application != defaults.MULTIMATIC:
            _LOGGER.warning(f"quick_mode is not compatible with {self._application}")
            return False
        await self._call_api(self.urls.system_quickmode, "delete")
        return True

    async def set_holiday_mode(self, start_date: date, end_date: date, temperature: float) -> None:
        """Set the :class:`~pymultimatic.model.mode.HolidayMode`.

        Args:
            start_date (date): Start date of the holiday mode.
            end_date (date): End date of the holiday mode.
            temperature (float): Target temperature while holiday mode
                :class:`~pymultimatic.model.mode.HolidayMode.is_applied`
        """
        payload = self.payloads.holiday_mode(True, start_date, end_date, self._round(temperature))
        await self._call_api(self.urls.system_holiday_mode, payload=payload)

    async def remove_holiday_mode(self) -> None:
        """Remove :class:`~pymultimatic.model.mode.HolidayMode`.

        Note:
            There is a little workaround here, since the API doesn't simply
            accept a DELETE request to remove the
            :class:`~pymultimatic.model.mode.HolidayMode`, so the manager is
            setting:

            * the start date to two days before

            * the end date to yesterday

            * temperature to frost protection

            * active to `False`

            This will ensure the
            :class:`~pymultimatic.model.mode.HolidayMode` is not active.

        """

        payload = self.payloads.holiday_mode(
            False,
            date.today() - timedelta(days=2),
            date.today() - timedelta(days=1),
            constants.FROST_PROTECTION_TEMP,
        )

        await self._call_api(self.urls.system_holiday_mode, payload=payload)

    async def set_hot_water_setpoint_temperature(self, dhw_id: str, temperature: float) -> None:
        """This set the target temperature for *hot water*."""
        _LOGGER.debug(f"Will set dhw target temperature to {temperature}")

        payload = self.payloads.hotwater_temperature_setpoint(self._round(temperature))

        await self._call_api(
            self.urls.hot_water_temperature_setpoint,
            params={"id": dhw_id},
            payload=payload,
        )

    async def set_hot_water_operating_mode(self, dhw_id: str, new_mode: OperatingMode) -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.HotWater`. The mode should be
        listed here :class:`~pymultimatic.model.component.HotWater.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Args:
            dhw_id (str): domestic hot water id.
            new_mode (OperatingMode): The new mode to set.
        """
        _LOGGER.debug(f"Will try to set hot water mode to {new_mode.name if new_mode else None}")

        if new_mode in HotWater.MODES:
            _LOGGER.debug(f"New mode is {new_mode}")
            await self._call_api(
                self.urls.hot_water_operating_mode,
                params={"id": dhw_id},
                payload=self.payloads.hot_water_operating_mode(new_mode.name),
            )
        else:
            _LOGGER.debug(
                f"Mode {new_mode.name if new_mode else None} is not available for hot water"
            )

    async def set_room_operating_mode(self, room_id: str, new_mode: OperatingMode) -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.Room`. The mode should be
        listed here :class:`~pymultimatic.model.component.Room.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Note:
            To set :class:`~pymultimatic.model.mode.QuickVeto`, you have to
            use :func:`set_room_quick_veto`

        Args:
            room_id (str): id of the room.
            new_mode (OperatingMode): The new mode to set.
        """
        if new_mode in Room.MODES and new_mode != OperatingModes.QUICK_VETO:
            _LOGGER.debug(f"New mode is {new_mode.name}")
            await self._call_api(
                self.urls.room_operating_mode,
                params={"id": room_id},
                payload=self.payloads.room_operating_mode(new_mode.name),
            )
        else:
            _LOGGER.debug(f"Mode {new_mode.name if new_mode else None} is not available for room")

    async def set_room_quick_veto(self, room_id: str, quick_veto: QuickVeto) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickVeto` for a
        :class:`~pymultimatic.model.component.Room`.
        It will override the current
        :class:`~pymultimatic.model.mode.QuickVeto`, if any.

        Args:
            room_id (str): Id of the room.
            quick_veto (QuickVeto): Quick veto to set.
        """
        payload = self.payloads.room_quick_veto(self._round(quick_veto.target), quick_veto.duration)
        await self._call_api(self.urls.room_quick_veto, params={"id": room_id}, payload=payload)

    async def remove_room_quick_veto(self, room_id: str) -> None:
        """Remove the :class:`~pymultimatic.model.mode.QuickVeto` from a
        :class:`~pymultimatic.model.component.Room`.

        Args:
            room_id (str): Id of the room.
        """

        await self._call_api(self.urls.room_quick_veto, "delete", params={"id": room_id})

    async def set_room_setpoint_temperature(self, room_id: str, temperature: float) -> None:
        """Set the new current target temperature for a
        :class:`~pymultimatic.model.component.Room`.

        Note:
            Setting the target temperature this way will actually set the room
            operating mode to
            :class:`~pymultimatic.model.mode.OperatingModes.MANUAL` (and
            target temperature to the desired temperature). It means the target
            temperature will remain the same until you change the target
            temperature again or the operating mode.

        Args:
            room_id (str): Id of the room.
            temperature (float): Target temperature to set.
        """

        _LOGGER.debug(f"Will try to set room target temperature to {temperature}")

        await self._call_api(
            self.urls.room_temperature_setpoint,
            params={"id": room_id},
            payload=self.payloads.room_temperature_setpoint(self._round(temperature)),
        )

    async def set_zone_quick_veto(self, zone_id: str, quick_veto: QuickVeto) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickVeto` for a
        :class:`~pymultimatic.model.component.Zone`.
        It will override the current
        :class:`~pymultimatic.model.mode.QuickVeto`, if any.

        Args:
            zone_id (str): Id of the zone.
            quick_veto (QuickVeto): Quick veto to set.
        """
        payload = self.payloads.zone_quick_veto(
            self._round(quick_veto.target),
            self._round(quick_veto.duration) if quick_veto.duration else None,
        )

        await self._call_api(self.urls.zone_quick_veto, params={"id": zone_id}, payload=payload)

    async def set_zone_heating_operating_mode(self, zone_id: str, new_mode: OperatingMode) -> None:
        """Set new operating mode to heat a
        :class:`~pymultimatic.model.component.Zone`. The mode should be
        listed here :class:`~pymultimatic.model.component.ZoneHeating.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Note:
            To set :class:`~pymultimatic.model.mode.QuickVeto`, you have to
            use :func:`set_zone_quick_veto`

        Args:
            zone_id (str): id of the zone.
            new_mode (OperatingMode): The new mode to set.
        """
        if zone_id and new_mode in ZoneHeating.MODES and new_mode != OperatingModes.QUICK_VETO:
            _LOGGER.debug(f"New mode is {new_mode}")
            await self._call_api(
                self.urls.zone_heating_mode,
                params={"id": zone_id},
                payload=self.payloads.zone_operating_mode(new_mode.name),
            )
        else:
            _LOGGER.debug(f"Mode {new_mode.name if new_mode else None} is not available for zone")

    async def set_zone_cooling_operating_mode(self, zone_id: str, new_mode: OperatingMode) -> None:
        """Set new operating mode to cool a
        :class:`~pymultimatic.model.component.Zone`. The mode should be
        listed here :class:`~pymultimatic.model.component.ZoneCooling.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Note:
            To set :class:`~pymultimatic.model.mode.QuickVeto`, you have to
            use :func:`set_zone_quick_veto`

        Args:
            zone_id (str): id of the zone.
            new_mode (OperatingMode): The new mode to set.
        """
        if new_mode in ZoneCooling.MODES and new_mode != OperatingModes.QUICK_VETO:
            _LOGGER.debug(f"New mode is {new_mode}")
            await self._call_api(
                self.urls.zone_cooling_mode,
                params={"id": zone_id},
                payload=self.payloads.zone_operating_mode(new_mode.name),
            )
        else:
            _LOGGER.debug(f"Mode {new_mode.name if new_mode else None} is not available for zone")

    async def remove_zone_quick_veto(self, zone_id: str) -> None:
        """Remove the :class:`~pymultimatic.model.mode.QuickVeto` from a
        :class:`~pymultimatic.model.component.Zone`.

        Args:
            zone_id (str): Id of the zone.
        """
        await self._call_api(self.urls.zone_quick_veto, "delete", params={"id": zone_id})

    async def set_zone_heating_setpoint_temperature(self, zone_id: str, temperature: float) -> None:
        """Set the configured temperature for the
        :class:`~pymultimatic.model.mode.SettingModes.DAY` mode.

        Note:
            It won't alter the
            :class:`~pymultimatic.model.mode.OperatingMode`.

        Args:
            zone_id (str): Id of the zone.
            temperature (float): New temperature.
        """
        _LOGGER.debug(f"Will try to set zone target temperature to {temperature}")

        payload = self.payloads.zone_temperature_setpoint(self._round(temperature))

        await self._call_api(
            self.urls.zone_heating_setpoint_temperature,
            params={"id": zone_id},
            payload=payload,
        )

    async def set_zone_cooling_setpoint_temperature(self, zone_id: str, temperature: float) -> None:
        """Set the configured cooling temperature for the
        :class:`~pymultimatic.model.mode.SettingModes.ON` mode.

        Note:
            It won't alter the
            :class:`~pymultimatic.model.mode.OperatingMode`.

        Args:
            zone_id (str): Id of the zone.
            temperature (float): New temperature.
        """
        _LOGGER.debug(f"Will try to set zone target temperature to {temperature}")

        payload = self.payloads.zone_temperature_setpoint(self._round(temperature))

        await self._call_api(
            self.urls.zone_cooling_setpoint_temperature,
            params={"id": zone_id},
            payload=payload,
        )

    async def set_zone_heating_setback_temperature(self, zone_id: str, temperature: float) -> None:
        """Set the configured heating temperature for the
        :class:`~pymultimatic.model.mode.SettingModes.NIGHT` mode.

        Note:
            It won't alter the
            :class:`~pymultimatic.model.mode.OperatingMode`.

        Args:
            zone_id (str): Id of the zone.
            temperature (float): New temperature.
        """
        _LOGGER.debug(f"Will try to set zone setback temperature to {temperature}")

        await self._call_api(
            self.urls.zone_heating_setback_temperature,
            params={"id": zone_id},
            payload=self.payloads.zone_temperature_setback(self._round(temperature)),
        )

    async def set_ventilation_operating_mode(
        self, ventilation_id: str, mode: OperatingMode
    ) -> None:
        """Set ventilation operating mode.
         Compatible modes are listed here
         :class:`~pymultimatic.model.Ventilation.MODES`

        Args:
            ventilation_id (str): id of the ventilation
            mode (OperatingMode): Mode to set
        """
        await self._call_api(
            self.urls.set_ventilation_operating_mode,
            params={"id": ventilation_id},
            payload=self.payloads.ventilation_operating_mode(mode.name),
        )

    async def set_ventilation_day_level(self, ventilation_id: str, level: int) -> None:
        """Set ventilation day level

        Args:
            ventilation_id (str): id of the ventilation
            level (int): Level between 1 and 6
        """
        await self._call_api(
            self.urls.set_ventilation_day_level,
            params={"id": ventilation_id},
            payload=self.payloads.ventilation_day_level(level),
        )

    async def set_ventilation_night_level(self, ventilation_id: str, level: int) -> None:
        """Set ventilation night level

        Args:
            ventilation_id (str): id of the ventilation
            level (int): Level between 1 and 6
        """
        await self._call_api(
            self.urls.set_ventilation_night_level,
            params={"id": ventilation_id},
            payload=self.payloads.ventilation_night_level(level),
        )

    async def request_hvac_update(self) -> None:
        """Request a hvac update. This allows the vaillant API to read the data from your system.

        This is necessary to update
        :class:`~pymultimatic.model.status.BoilerStatus` and
        :class:`~pymultimatic.model.status.Error`

        Note:
            The **request** done by the manager is done **synchronously**,
            but the **update** processing (basically, reading data from your
            system) is done **asynchronously** by vaillant API.

        Note:
            The manager will try to reduce number of call, I think the call is
            costly. If the
            :class:`~pymultimatic.model.syncstate.SyncState`
            :class:`~pymultimatic.model.syncstate.SyncState.is_pending`,
            the manager will skip the call to the API.


        it can take some times for the update to occur (Most of the
        time, it takes about 1 or 2 minutes before you can see changes, if any)

        It the request is done too often, the API may return an error
        (HTTP 409).

        """

        state = mapper.map_hvac_sync_state(await self._call_api(self.urls.hvac))

        if state and not state.is_pending:
            await self._call_api(self.urls.hvac_update, "put")

    async def set_datetime(self, dt: datetime) -> None:
        """Sets the system datetime

        Args:
            dt (datetime): the datetime to set
        """
        await self._call_api(
            self.urls.system_datetime,
            "put",
            payload={"datetime": dt.isoformat(timespec="microseconds")},
        )

    @staticmethod
    def _round(number: float) -> float:
        """round a float to the nearest 0.5, as vaillant API only accepts 0.5
        stepself."""
        return round(number * 2) / 2

    @retry_async(
        on_exceptions=(WrongResponseError,),
        on_status_codes=tuple(range(500, 600)),
        backoff_base=1,
        num_tries=3,
    )
    async def _call_api(
        self,
        url_call: Callable[..., str],
        method: Optional[str] = None,
        schema: Optional[Schema] = None,
        **kwargs: Any,
    ) -> Any:
        await self._ensure_ready()

        params = kwargs.get("params", {})
        params.update({"serial": self._serial})

        payload = kwargs.get("payload", None)

        if method is None:
            method = "get"
            if payload is not None:
                method = "put"

        url = url_call(**params)
        response = await self._connector.request(method, url, payload)
        if schema:
            return await self._validate_schema(schema, response, url)
        return response

    @staticmethod
    async def _validate_schema(schema: Schema, response: Any, url: str) -> Any:
        try:
            return schema.validate(response)
        except SchemaError as err:
            raise WrongResponseError(
                message=f"Cannot validate response from {url}: {err.code}",
                response=response,
                status=200,
            ) from err

    async def _ensure_ready(self) -> None:
        if not await self._connector.is_logged():
            async with self._ensure_ready_lock:
                # double check whether other coroutine has already logged in
                if not await self._connector.is_logged():
                    await self._connector.login()
                    await self._fetch_serial()
        if not self._serial:
            async with self._ensure_ready_lock:
                if not self._serial:
                    await self._fetch_serial()

    async def _fetch_serial(self) -> None:
        if not self._fixed_serial:
            facilities = await self._connector.get(self.urls.facilities_list())
            self._serial = mapper.map_serial_number(facilities)



"""Common structure."""

import abc
from datetime import datetime
from typing import Optional

import attr

from . import ActiveMode, OperatingMode, OperatingModes, QuickVeto, SettingModes, TimeProgram


@attr.s
class Function:
    """This is a common class for function in the system. A function is
    basically something that has a time program and an operating mode.

    It allows you to get the :class:`~pymultimatic.model.mode.ActiveMode`.

    Args:
        time_program (TimeProgram): Used for time based operating mode,
            :class:`~pymultimatic.model.mode.OperatingModes.AUTO`
        operating_mode (OperatingMode): Configured operating mode.
        target_high (float): Temperature the component is trying to
            reach, `None` for :class:`Circulation`.
        target_low: (float): Target low temperature of the component.

    """

    time_program = attr.ib(type=TimeProgram, default=None)
    operating_mode = attr.ib(type=OperatingMode, default=None)
    target_high = attr.ib(type=Optional[float], default=None)
    target_low = attr.ib(type=Optional[float], default=None)

    @property
    def active_mode(self) -> ActiveMode:
        """ActiveMode: Get the :class:`~pymultimatic.model.mode.ActiveMode`
        for this function. All operating modes are handled,
        **but not quick veto nor quick mode.**

        Note:
        :class:`Function.active_mode` should not be used directly through the
        function, actually, the function is only aware of itself, so in case
        of a :class:`~pymultimatic.model.mode.QuickMode` or
        :class:`~pymultimatic.model.mode.QuickVeto`  is running on, you
        may not receive the active mode that is really applied. You should use
        :class:`~pymultimatic.model.system.System` for that.
        """
        mode = None
        if (
            self.operating_mode == OperatingModes.AUTO
            or self.operating_mode == OperatingModes.TIME_CONTROLLED
        ):
            if self.time_program:
                setting = self.time_program.get_for(datetime.now())
                if setting:
                    if setting.setting in [SettingModes.DAY, SettingModes.ON]:
                        if self.operating_mode == OperatingModes.AUTO:
                            target = self.target_high
                        else:
                            target = setting.target_temperature
                        mode = ActiveMode(target, self.operating_mode, setting.setting)
                    else:
                        mode = ActiveMode(self.target_low, self.operating_mode, setting.setting)
        if not mode:
            mode = self._active_mode()
        return mode

    @abc.abstractmethod
    def _active_mode(self) -> ActiveMode:
        pass


@attr.s
class Component:
    """This is a common class for components in the system.
    A component can be :class:`~pymultimatic.model.room.Room`,
    :class:`~pymultimatic.model.zone.Zone`,
    :class:`~pymultimatic.model.dhw.HotWater` or
    :class:`~pymultimatic.model.dhw.Circulation`.


    It allows you to get the :class:`~pymultimatic.model.mode.ActiveMode`.

    Note:
        :class:`active_mode` should not be used directly through the component,
        actually, the component is only aware of itself, so in case of a
        :class:`~pymultimatic.model.mode.QuickMode` is running on, you may
        not receive the active mode that is really applied. You should use
        :class:`~pymultimatic.model.system.System` for that.

    Args:
        id (str): Id of the component, this is used to get data from the API.
        name (str): Name of the component, this is set by the user through the
            mobile app or via the VRC700.
        temperature (float): Current temperature for the component,
            this is `None` for :class:`Circulation`
        quick_veto (QuickVeto): Will be populated if there is a
            :class:`~pymultimatic.model.mode.QuickVeto` running on. Always
            `None` for :class:`Circulation` and :class:`HotWater`.
    """

    id = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    temperature = attr.ib(type=Optional[float], default=None)
    quick_veto = attr.ib(type=Optional[QuickVeto], default=None)

    @property
    @abc.abstractmethod
    def active_mode(self) -> Optional[ActiveMode]:
        """ActiveMode: Get the :class:`~pymultimatic.model.mode.ActiveMode`
        for this component. All operating modes and quick veto are handled,
        **but not quick mode.**

        Note:
        :class:`Component.active_mode` should not be used directly through the
        component, actually, the component is only aware of itself, so in case
        of a :class:`~pymultimatic.model.mode.QuickMode` is running on, you
        may not receive the active mode that is really applied. You should use
        :class:`~pymultimatic.model.system.System` for that.
        """



"""Test for payloads"""
import inspect
import json
import unittest
from datetime import date, datetime
from typing import Any, Dict, Union

from pymultimatic.api import payloads, payloads_senso


class PayloadsTest(unittest.TestCase):
    """Test class."""

    def test_all_payload(self) -> None:
        """Test that ensure all payload are well json formatted."""
        functions_list = inspect.getmembers(payloads, predicate=inspect.isfunction)

        self.assertTrue(len(functions_list) > 0)

        for function in functions_list:
            args = self._get_args_name(function[1])

            if not args:
                url = function[1]()
                self._assert_function_call(url)
            else:
                payload = function[1](**args)
                self._assert_function_call(payload)

        payload = payloads.room_quick_veto(15, None)
        self._assert_function_call(payload)
        assert payload["duration"] == 180

    def test_all_payload_senso(self) -> None:
        """Test that ensure all payload senso are well json formatted."""
        functions_list = inspect.getmembers(payloads_senso, predicate=inspect.isfunction)

        self.assertTrue(len(functions_list) > 0)

        for function in functions_list:
            args = self._get_args_name(function[1])

            if not args:
                url = function[1]()
                self._assert_function_call(url)
            else:
                payload = function[1](**args)
                self._assert_function_call(payload)

        payload = payloads_senso.room_quick_veto(15, None)
        self._assert_function_call(payload)
        assert payload["duration"] == 180

    def _get_args_name(self, function: Any) -> Dict[str, Any]:
        args: Dict[str, Any] = {}
        params = inspect.signature(function).parameters

        for item in inspect.signature(function).parameters.items():
            cls = params[item[0]].annotation

            if cls == bool:
                args[item[0]] = False
            elif cls == float:
                args[item[0]] = 10.0
            elif cls == Union[float, None]:
                args[item[0]] = 10.0
            elif cls == date:
                args[item[0]] = datetime.now()
            elif cls == str:
                args[item[0]] = "test"
            elif cls == int:
                args[item[0]] = 10
            elif cls == Union[int, None]:
                args[item[0]] = 10
            else:
                self.fail("Unhandled class " + cls.__name__)

        return args

    def _assert_function_call(self, result: Any) -> Any:
        json.loads(json.dumps(result))



"""Groups all models related to DHW."""
from typing import Optional

import attr

from . import ActiveMode, Component, Function, OperatingModes, SettingModes, constants


@attr.s
class HotWater(Function, Component):
    """This is representing the hot water from the system.

    There is no `quick_veto` available for this component.
    """

    MODES = [OperatingModes.ON, OperatingModes.OFF, OperatingModes.AUTO]
    """List of mode that are applicable to the hot water component."""

    MIN_TARGET_TEMP = 35
    """Min `target temperature` for the hot water."""

    MAX_TARGET_TEMP = 70
    """Max `target temperature` for the hot water."""

    quick_veto = attr.ib(default=None, init=False)
    target_low = attr.ib(default=MIN_TARGET_TEMP, init=False)

    def _active_mode(self) -> ActiveMode:
        mode: ActiveMode
        if (
            self.operating_mode == OperatingModes.AUTO
        ):  # auto at this point mean we have a "direct heater"
            mode = ActiveMode(self.target_high, OperatingModes.AUTO, SettingModes.ON)
        elif (
            self.operating_mode == OperatingModes.ON or self.operating_mode == OperatingModes.MANUAL
        ):
            mode = ActiveMode(self.target_high, self.operating_mode)
        else:  # MODE_OFF
            mode = ActiveMode(constants.FROST_PROTECTION_TEMP, OperatingModes.OFF)
        return mode


@attr.s
class Circulation(Function, Component):
    """This is representing the circulation from the system.
    This is a bit special component since there is no
    `current_temperature`, `target_temperature` nor `quick_veto`.
    """

    MODES = [OperatingModes.ON, OperatingModes.OFF, OperatingModes.AUTO]
    """List of mode that are applicable to the hot water component."""

    temperature = attr.ib(default=None, init=False)
    quick_veto = attr.ib(default=None, init=False)
    target_high = attr.ib(default=None, init=False)
    target_low = attr.ib(default=None, init=False)

    def _active_mode(self) -> ActiveMode:
        mode: ActiveMode
        if (
            self.operating_mode == OperatingModes.AUTO
        ):  # auto at this point mean we have a "direct heater"
            mode = ActiveMode(self.target_low, OperatingModes.AUTO, SettingModes.OFF)
        elif self.operating_mode == OperatingModes.ON:
            mode = ActiveMode(self.target_high, OperatingModes.ON)
        else:  # MODE_OFF
            mode = ActiveMode(self.target_low, OperatingModes.OFF)
        return mode


@attr.s
class Dhw:
    """This is representing the DHW (Domestic Hot Water) from the system."""

    hotwater = attr.ib(type=Optional[HotWater], default=None)
    circulation = attr.ib(type=Optional[Circulation], default=None)



import inspect
import unittest

from pymultimatic.api import urls, urls_senso


class TestUrls(unittest.TestCase):
    def test_all_urls(self) -> None:
        functions_list = inspect.getmembers(urls, predicate=inspect.isroutine)

        self.assertTrue(len(functions_list) > 0)

        not_called = 0
        for function in functions_list:
            params = inspect.signature(function[1]).parameters.items()
            values = {
                "serial": "123",
                "id": "456",
                "sgtin": "789",
                "device_id": "111",
                "report_id": "222",
            }
            if len(params) > 1:
                values.update(
                    {
                        "energy_type": "type",
                        "function": "func",
                        "time_range": "range",
                        "start": "start",
                        "offset": "offset",
                    }
                )
                not_called += 1

            function[1](**values)

        assert not_called == 1

    def test_all_urls_senso(self) -> None:
        functions_list = inspect.getmembers(urls_senso, predicate=inspect.isroutine)

        self.assertTrue(len(functions_list) > 0)

        not_called = 0
        for function in functions_list:
            params = inspect.signature(function[1]).parameters.items()
            values = {
                "serial": "123",
                "id": "456",
                "sgtin": "789",
                "device_id": "111",
                "report_id": "222",
            }
            if len(params) > 1:
                values.update(
                    {
                        "energy_type": "type",
                        "function": "func",
                        "time_range": "range",
                        "start": "start",
                        "offset": "offset",
                    }
                )
                not_called += 1

            function[1](**values)

        assert not_called == 1



"""Groups all API payloads for MULTIMATIC. Payload are always json formatted."""

from datetime import date
from typing import Any, Dict, Optional

_DATE_FORMAT = "%Y-%m-%d"


def hotwater_temperature_setpoint(temperature: float) -> Dict[str, Any]:
    """Payload used to set target temperature for
    :class:`~pymultimatic.model.component.HotWater`.
    """
    return {"temperature_setpoint": temperature}


def room_temperature_setpoint(temperature: float) -> Dict[str, Any]:
    """Payload used to set target temperature for
    :class:`~pymultimatic.model.component.Room`.
    """
    return {"temperatureSetpoint": temperature}


def zone_temperature_setpoint(temperature: float) -> Dict[str, Any]:
    """Payload used to set target temperature for
    :class:`~pymultimatic.model.component.Zone`.
    """
    return {"setpoint_temperature": temperature}


def zone_temperature_setback(temperature: float) -> Dict[str, Any]:
    """Payload used to set setback temperature for
    :class:`~pymultimatic.model.component.Zone`.
    """
    return {"setback_temperature": temperature}


def hot_water_operating_mode(mode: str) -> Dict[str, Any]:
    """Payload to set operating mode for
    :class:`~pymultimatic.model.component.HotWater`.
    """
    return {"operation_mode": mode}


def room_operating_mode(mode: str) -> Dict[str, Any]:
    """Payload to set operating mode for
    :class:`~pymultimatic.model.component.Room`.
    """
    return {"operationMode": mode}


def zone_operating_mode(mode: str) -> Dict[str, Any]:
    """Payload to set operating mode for
    :class:`~pymultimatic.model.component.Zone`.
    """
    return {"mode": mode}


def quickmode(quick_mode: str, duration: Optional[int] = None) -> Dict[str, Any]:
    """Payload to set :class:`~pymultimatic.model.mode.QuickMode` for the
    system.

    Duration is mandatory (Duration is in minutes, max 1440 =24 hours).
    """
    payload: Dict[str, Any] = {"quickmode": {"quickmode": quick_mode}}

    if duration:
        payload["quickmode"].update({"duration": duration})

    return payload


def zone_quick_veto(temperature: float, duration: Optional[float] = 6) -> Dict[str, Any]:
    """Payload to set a :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Zone`.

    The duration is not configurable by the API, it's 6 hours
    """
    return {"setpoint_temperature": temperature}


def room_quick_veto(temperature: float, duration: Optional[int]) -> Dict[str, Any]:
    """Payload to set a :class:`~pymultimatic.model.mode.QuickVeto` for a
    :class:`~pymultimatic.model.component.Room`.

    Duration is mandatory (Duration is in minutes, max 1440 =24 hours).
    """

    if not duration:
        duration = 180

    return {"temperatureSetpoint": temperature, "duration": duration}


def holiday_mode(
    active: bool, start_date: date, end_date: date, temperature: float
) -> Dict[str, Any]:
    """Payload to set :class:`~pymultimatic.model.mode.HolidayMode`."""
    return {
        "active": active,
        "start_date": start_date.strftime(_DATE_FORMAT),
        "end_date": end_date.strftime(_DATE_FORMAT),
        "temperature_setpoint": temperature,
    }


def ventilation_operating_mode(mode: str) -> Dict[str, Any]:
    """Payload to set operating mode for
    :class:`~pymultimatic.model.Ventilation`.
    """
    return {"mode": mode}


def ventilation_level(level: int) -> Dict[str, Any]:
    """Payload to set level for
    :class:`~pymultimatic.model.Ventilation`.
    """
    return {"level": level}



"""Tests for circulation."""
import unittest

from pymultimatic.model import OperatingModes
from tests.conftest import _circulation


class CirculationTest(unittest.TestCase):
    """Test class."""

    def test_get_active_mode_on(self) -> None:
        """Get active mode when operation mode is ON."""
        circulation = _circulation()
        circulation.operating_mode = OperatingModes.ON

        active_mode = circulation.active_mode

        self.assertEqual(OperatingModes.ON, active_mode.current)
        self.assertIsNone(active_mode.target)
        self.assertIsNone(active_mode.sub)

    def test_get_active_mode_off(self) -> None:
        """Get active mode when operation mode is OFF."""
        circulation = _circulation()
        circulation.operating_mode = OperatingModes.OFF

        active_mode = circulation.active_mode

        self.assertEqual(OperatingModes.OFF, active_mode.current)
        self.assertIsNone(active_mode.target)
        self.assertIsNone(active_mode.sub)



from setuptools import setup, find_packages

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pymultiMATIC",
    version="0.7.2",
    description="Python interface with Vaillant multiMATIC",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/thomasgermain/pymultiMATIC.git",
    author="Thomas Germain",
    author_email="12560542+thomasgermain@users.noreply.github.com",
    license="MIT",
    packages=find_packages(exclude=("tests", "tests/*", "/tests", "/tests/*")),
    zip_safe=False,
    setup_requires=["pytest-runner"],
    install_requires=[
        "attrs>=23.0.0,<24.0.0",
        "aiohttp>=3.8.0,<4.0.0",
        "schema>=0.7.4,<0.8.0",
    ],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Home Automation",
    ],
)



"""Information coming from API."""
import attr


@attr.s
class FacilityDetail:
    """Facility information.

    Args:
        name (str): Name of the facility.
        serial_number (str): Serial number of the facility.
        firmware_version (str): Version of the firmware of the facility.
        ethernet_mac (str): Ethernet mac address
        wifi_mac (str): Wifi mac address
    """

    name = attr.ib(type=str)
    serial_number = attr.ib(type=str)
    firmware_version = attr.ib(type=str)
    ethernet_mac = attr.ib(type=str)
    wifi_mac = attr.ib(type=str)



"""Expected API response schemas"""
from schema import And, Optional, Or, Schema

non_empty_str = And(str, len)  # pylint: disable=invalid-name
optional_str = Or(str, None)
numeric = Or(int, float)  # pylint: disable=invalid-name
not_rbr = And(non_empty_str, lambda v: v not in ("RBR", "rbr"))  # pylint: disable=invalid-name

FACILITIES = Schema(
    {
        "body": {
            "facilitiesList": [
                {
                    "serialNumber": non_empty_str,
                    "name": non_empty_str,
                    Optional("responsibleCountryCode"): non_empty_str,
                    Optional("supportedBrand"): non_empty_str,
                    "firmwareVersion": non_empty_str,
                    Optional("capabilities"): [non_empty_str],
                    "networkInformation": {
                        "macAddressEthernet": non_empty_str,
                        Optional("macAddressWifiAccessPoint"): non_empty_str,
                        Optional("macAddressWifiClient"): non_empty_str,
                    },
                }
            ]
        },
    },
    ignore_extra_keys=True,
)

# could have used calendar.day_name, but locale may cause issues on different systems
TIMEPROGRAM_PART = Schema(
    {
        day: [
            {
                Or("startTime", "start_time"): non_empty_str,  # TODO: parse time
                Optional(Or("temperatureSetpoint", "setpoint")): numeric,
                Optional("setting"): non_empty_str,  # TODO: ENUM
                Optional("mode"): non_empty_str,  # TODO: ENUM
                Optional("end_time"): non_empty_str,  # TODO: parse time
            }
        ]
        for day in (
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        )
    },
    ignore_extra_keys=True,
)

CONTROLLED_BY = Schema(
    {
        "name": not_rbr,
        Optional("link"): {
            Optional("rel"): str,
            Optional("resourceLink"): str,
            Optional("name"): str,
        },
    }
)

_CONTROLLED_BY_RBR = Schema(
    {
        "name": Or("RBR", "rbr"),
        Optional("link"): {
            Optional("rel"): str,
            Optional("resourceLink"): str,
            Optional("name"): str,
        },
    }
)

_CONTROLLED_BY_QUICK_VETO = Schema(
    {
        "name": "QUICK_VETO",
        Optional("link"): {
            Optional("rel"): str,
            Optional("resourceLink"): str,
            Optional("name"): str,
        },
    }
)


_OPTIONAL_CONFIG_FUNCTION_PART = Schema(
    {
        Optional("configuration"): {
            Or("mode", "operation_mode", "operationMode"): non_empty_str,  # TODO: ENUM
            Optional(
                Or(
                    "setpoint_temperature",
                    "temperature_setpoint",
                    "temperatureSetpoint",
                    "day_level",
                    "hotwater_temperature_setpoint",
                    "manual_mode_temperature_setpoint",
                )
            ): numeric,
            Optional(
                Or("setback_temperature", "night_level", "setback_temperature_setpoint")
            ): numeric,
            Optional("day_level"): int,
            Optional("night_level"): int,
        },
        Optional("timeprogram"): TIMEPROGRAM_PART,
    },
    ignore_extra_keys=True,
)

HOT_WATER = Schema({"body": _OPTIONAL_CONFIG_FUNCTION_PART}, ignore_extra_keys=True)

FUNCTION_PART = Schema(
    {
        "configuration": {
            Or("mode", "operation_mode", "operationMode"): non_empty_str,  # TODO: ENUM
            Optional(
                Or(
                    "setpoint_temperature",
                    "temperature_setpoint",
                    "temperatureSetpoint",
                    "day_level",
                    "manual_mode_temperature_setpoint",
                )
            ): numeric,
            Optional(
                Or("setback_temperature", "night_level", "setback_temperature_setpoint")
            ): numeric,
            Optional("day_level"): int,
            Optional("night_level"): int,
        },
        "timeprogram": TIMEPROGRAM_PART,
    },
    ignore_extra_keys=True,
)

QUICK_MODE = Schema(
    {
        Optional("quickmode"): non_empty_str,  # TODO: ENUM
        Optional("duration"): numeric,
    }
)

_CONFIGURATION_AWAY = Schema(
    {
        "start_datetime": non_empty_str,
        "end_datetime": non_empty_str,
        Optional("temperature_setpoint"): numeric,
    },
)

_ZONE_CONFIGURATION = Schema(
    {
        "name": non_empty_str,
        Optional("enabled"): bool,
        Optional("inside_temperature"): numeric,
        Optional("active_function"): non_empty_str,  # TODO: add ENUM validation
        Optional("current_quickmode"): non_empty_str,
        Optional("quick_veto"): Or(
            {
                "active": bool,
                "setpoint_temperature": numeric,
            },
            {
                Optional("expires_at"): str,
                "temperature_setpoint": numeric,
            },
        ),
        Optional("quickmode"): QUICK_MODE,
        Optional("eco_mode"): bool,
        Optional("current_desired_setpoint"): numeric,
        Optional("away"): _CONFIGURATION_AWAY,
        Optional("manual_cooling_active"): bool,
        Optional("humidity"): numeric,
    }
)

_DISABLED_ZONE_CONFIGURATION = Schema(
    {
        Optional("name"): non_empty_str,
        "enabled": False,
        Optional("inside_temperature"): numeric,
        Optional("active_function"): non_empty_str,  # TODO: add ENUM validation
        Optional("quick_veto"): {
            "active": bool,
            "setpoint_temperature": numeric,
        },
        Optional("quickmode"): QUICK_MODE,
    }
)


_ZONE_PART = Schema(
    Or(
        Schema(  # most common zone response
            {
                "_id": non_empty_str,
                "configuration": _ZONE_CONFIGURATION,
                Optional("currently_controlled_by"): CONTROLLED_BY,
                Optional("heating"): FUNCTION_PART,
                Optional("cooling"): FUNCTION_PART,
            }
        ),
        Schema(  # disabled zone, in that case, almost everything is optional
            {
                "_id": non_empty_str,
                "configuration": _DISABLED_ZONE_CONFIGURATION,
                Optional("currently_controlled_by"): CONTROLLED_BY,
                Optional("heating"): _OPTIONAL_CONFIG_FUNCTION_PART,
                Optional("cooling"): _OPTIONAL_CONFIG_FUNCTION_PART,
            }
        ),
        Schema(  # rbr zone, in that case, almost everything is optional
            {
                "_id": non_empty_str,
                Optional("configuration"): _ZONE_CONFIGURATION,
                Optional("currently_controlled_by"): _CONTROLLED_BY_RBR,
                Optional("heating"): _OPTIONAL_CONFIG_FUNCTION_PART,
                Optional("cooling"): _OPTIONAL_CONFIG_FUNCTION_PART,
            }
        ),
        Schema(  # rbr zone vr920, in that case, almost everything is optional
            {
                "_id": non_empty_str,
                Optional("configuration"): _ZONE_CONFIGURATION,
                Optional("currently_controlled_by"): non_empty_str,
                Optional("heating"): _OPTIONAL_CONFIG_FUNCTION_PART,
                Optional("cooling"): _OPTIONAL_CONFIG_FUNCTION_PART,
            }
        ),
        Schema(  # quick veto set on a zone, configuration of heating (or cooling) is optional
            {
                "_id": non_empty_str,
                "configuration": _ZONE_CONFIGURATION,
                Optional("currently_controlled_by"): _CONTROLLED_BY_QUICK_VETO,
                Optional("heating"): _OPTIONAL_CONFIG_FUNCTION_PART,
                Optional("cooling"): _OPTIONAL_CONFIG_FUNCTION_PART,
            }
        ),
    ),
    ignore_extra_keys=True,
)

_DHW_SENSO_PART = Schema(
    {
        "hotwater": _OPTIONAL_CONFIG_FUNCTION_PART,
        "circulation": _OPTIONAL_CONFIG_FUNCTION_PART,
        Optional("controlled_by"): CONTROLLED_BY,
        Optional("configuration"): {
            Optional("away"): _CONFIGURATION_AWAY,
        },
    }
)

_DHW_PART = Schema(
    {
        "_id": non_empty_str,
        "hotwater": _OPTIONAL_CONFIG_FUNCTION_PART,
        "circulation": _OPTIONAL_CONFIG_FUNCTION_PART,
        Optional("controlled_by"): CONTROLLED_BY,
    },
    ignore_extra_keys=True,
)

DHWS = Schema({"body": Or([_DHW_PART], _DHW_SENSO_PART)}, ignore_extra_keys=True)

_VENTILATION_PART = Schema(
    {
        "_id": non_empty_str,
        "fan": FUNCTION_PART,
    },
    ignore_extra_keys=True,
)

VENTILATION_LIST = Schema({"body": [_VENTILATION_PART]}, ignore_extra_keys=True)

VENTILATION = Schema({"body": _VENTILATION_PART}, ignore_extra_keys=True)

SYSTEM = Schema(
    {
        "body": {
            "configuration": {
                Optional("eco_mode"): bool,
                Optional("holidaymode"): {
                    "active": bool,
                    "start_date": non_empty_str,  # TODO: parse date
                    "end_date": non_empty_str,  # TODO: parse date
                    "temperature_setpoint": numeric,
                },
                Optional("manual_cooling"): {
                    "start_date": non_empty_str,  # TODO: parse date
                    "end_date": non_empty_str,  # TODO: parse date
                },
                Optional("quickmode"): QUICK_MODE,
            },
            "status": {
                "datetime": non_empty_str,  # TODO: parse date
                Optional("outside_temperature"): numeric,
            },
            "zones": [_ZONE_PART],
            Optional("dhw"): Or([_DHW_PART], _DHW_SENSO_PART),
            Optional("ventilation"): [_VENTILATION_PART],
        }
    },
    ignore_extra_keys=True,
)


FUNCTION = Schema(
    {
        "body": FUNCTION_PART,
    },
    ignore_extra_keys=True,
)

ROOM_PART = Schema(
    {
        "roomIndex": int,
        "timeprogram": TIMEPROGRAM_PART,
        "configuration": {
            "name": non_empty_str,
            "temperatureSetpoint": numeric,
            "operationMode": non_empty_str,  # TODO: ENUM
            "currentTemperature": numeric,
            "childLock": bool,
            "isWindowOpen": bool,
            Optional("iconId"): str,
            Optional("currentHumidity"): numeric,
            "devices": [
                {
                    "name": optional_str,
                    "sgtin": non_empty_str,
                    "deviceType": non_empty_str,  # TODO: ENUM
                    "isBatteryLow": bool,
                    "isRadioOutOfReach": bool,
                }
            ],
            Optional("quickVeto"): {
                "remainingDuration": numeric,
            },
        },
    },
    ignore_extra_keys=True,
)

ROOM = Schema(
    {
        "body": ROOM_PART,
    },
    ignore_extra_keys=True,
)

ROOM_LIST = Schema(
    {
        "body": {
            "rooms": [ROOM_PART],
        },
    },
    ignore_extra_keys=True,
)

ZONE = Schema(
    {
        "body": _ZONE_PART,
    },
    ignore_extra_keys=True,
)

ZONE_LIST = Schema(
    {
        "body": [_ZONE_PART],
    },
    ignore_extra_keys=True,
)

GATEWAY = Schema(
    {
        "body": {
            "gatewayType": non_empty_str,
        },
    },
    ignore_extra_keys=True,
)

HVAC = Schema(
    {
        "meta": {
            "onlineStatus": {
                "status": non_empty_str,  # TODO: ENUM
            },
            "firmwareUpdateStatus": {
                "status": non_empty_str,  # TODO: ENUM
            },
        },
        "body": {
            Optional("errorMessages"): [
                {
                    "type": non_empty_str,  # TODO: ENUM
                    "timestamp": int,  # TODO: parse timestamp
                    "deviceName": non_empty_str,
                    "statusCode": non_empty_str,
                    "title": non_empty_str,
                    "description": non_empty_str,
                    Optional("hint"): non_empty_str,
                }
            ]
        },
    },
    ignore_extra_keys=True,
)

_LIVE_REPORT_PART = Schema(
    {
        "_id": non_empty_str,
        "name": non_empty_str,
        "value": numeric,
        "unit": non_empty_str,
        "measurement_category": non_empty_str,  # TODO: ENUM
        Optional("associated_device_function"): non_empty_str,  # TODO: ENUM
    },
    ignore_extra_keys=True,
)

LIVE_REPORT = Schema({"body": _LIVE_REPORT_PART}, ignore_extra_keys=True)

LIVE_REPORTS = Schema(
    {
        "body": {
            "devices": [
                {
                    "_id": non_empty_str,
                    "name": non_empty_str,
                    "reports": [_LIVE_REPORT_PART],
                }
            ],
        },
    },
    ignore_extra_keys=True,
)



"""Tests schema."""
import json
import unittest

from schema import SchemaError

from pymultimatic.api import schemas
from tests.conftest import path, senso_responses_folders


class SchemaTest(unittest.TestCase):
    """Test class."""

    def test_schema_system_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = [
            "systemcontrol",
            "systemcontrol_holiday",
            "systemcontrol_hotwater_boost",
            "systemcontrol_no_outside_temp",
            "systemcontrol_off",
            "systemcontrol_quick_veto",
            "systemcontrol_ventilation",
            "systemcontrol_zone_no_config_rbr",
        ]
        for i in range(len(files)):
            files[i] = my_path + files[i]

        for folder in senso_responses_folders():
            files.append(folder + "/system")

        for file in files:
            with open(path(file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.SYSTEM.validate(json_val)
                json_val.pop("meta")

                body = json_val.get("body")
                if "parameters" in body:
                    body.pop("parameters")
                self.maxDiff = None
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_system_validation_error(self) -> None:
        """Ensure validation fails."""
        with open(path("files/responses/systemcontrol_zone_no_config"), "r") as open_f:
            json_val = json.loads(open_f.read())
            try:
                schemas.SYSTEM.validate(json_val)
            except SchemaError as err:
                self.assertIn("Missing key: 'configuration'", err.args[0])

    def test_schema_livereport_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"
        files = [
            "livereport",
            "livereport_FlowTemperatureVF1",
        ]
        for i in range(len(files)):
            files[i] = my_path + files[i]

        for folder in senso_responses_folders():
            files.append(folder + "/live_report")

        for file in files:
            with open(path(file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.LIVE_REPORTS.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_livereport_single_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"
        files = ["livereport_single"]

        for file in files:
            with open(path(my_path + file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.LIVE_REPORT.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_hvac_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = [
            "hvacstate",
            "hvacstate_empty",
            "hvacstate_errors",
            "hvacstate_pending",
        ]
        for i in range(len(files)):
            files[i] = my_path + files[i]

        for folder in senso_responses_folders():
            files.append(folder + "/hvac")

        for file in files:
            with open(path(file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.HVAC.validate(json_val)
                json_val.get("meta").pop("syncState")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_facilities_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = [
            "facilities",
            "facilities_multiple",
        ]
        for i in range(len(files)):
            files[i] = my_path + files[i]

        for folder in senso_responses_folders():
            files.append(folder + "/facilities_list")

        for file in files:
            with open(path(file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.FACILITIES.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_zone_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = [
            "zone",
            "zone_always_off",
            "zone_always_on",
            "zone_no_active_function",
            "zone_no_quickveto",
        ]

        for file in files:
            with open(path(my_path + file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.ZONE.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_zones_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = [
            "zones",
            "zones_3_zones",
            "zones_missing_heating_config_quick_veto",
        ]
        for i in range(len(files)):
            files[i] = my_path + files[i]

        for folder in senso_responses_folders():
            files.append(folder + "/zones")
            files.append(folder + "/zones_manual")

        for file in files:
            with open(path(file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.ZONE_LIST.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_rooms_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = ["rooms", "rooms_quick_veto"]

        for file in files:
            with open(path(my_path + file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.ROOM_LIST.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_room_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = ["room", "room_empty_device_name"]

        for file in files:
            with open(path(my_path + file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.ROOM.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_dhw_validation(self) -> None:
        """Ensure schema validation doesn't alter the response"""
        my_path = "files/responses/"

        files = [
            "dhws",
            "dhws_minimal",
        ]
        for i in range(len(files)):
            files[i] = my_path + files[i]

        for folder in senso_responses_folders():
            files.append(folder + "/dhw")
            files.append(folder + "/dhws")

        for file in files:
            with open(path(file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.DHWS.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)

    def test_schema_validation(self) -> None:
        """Ensure validation works"""
        my_path = "files/responses/"
        files = ["hotwater", "hotwater_always_off", "hotwater_always_on"]

        for file in files:
            with open(path(my_path + file), "r") as open_f:
                json_val = json.loads(open_f.read())
                result = schemas.HOT_WATER.validate(json_val)
                json_val.pop("meta")
                self.assertDictEqual(result, json_val, "error for " + file)



"""Utility tool."""
from datetime import datetime, timedelta
from typing import Callable, Dict, Optional

from pymultimatic.model import (
    ActiveMode,
    Circulation,
    Component,
    HolidayMode,
    HotWater,
    QuickMode,
    QuickModes,
    Room,
    SettingModes,
    TimePeriodSetting,
    TimeProgram,
    Ventilation,
    Zone,
    constants,
)


def active_mode_for(
    comp: Optional[Component], holiday: HolidayMode, quick_mode: Optional[QuickMode]
) -> Optional[ActiveMode]:
    """Get active mode for a given component"""
    if comp is not None:
        return _method_by_component[comp.__class__.__name__](comp, holiday, quick_mode)
    return None


def _active_mode_for_zone(
    zone: Zone, holiday: HolidayMode, quick_mode: QuickMode
) -> Optional[ActiveMode]:
    """Get the current
    :class:`~pymultimatic.model.mode.ActiveMode` for a
    :class:`~pymultimatic.model.component.Zone`. This is the only way to
    get the real one.

    Args:
        zone (Zone): The zone you want the active mode for.

    Returns:
        ActiveMode: The active mode.
    """
    mode: Optional[ActiveMode] = zone.active_mode

    # Holiday mode takes precedence over everything
    if holiday and holiday.active_mode:
        mode = holiday.active_mode

    # Global system quick mode takes over zone settings
    if quick_mode and quick_mode.for_zone:
        if quick_mode in (
            QuickModes.VENTILATION_BOOST,
            QuickModes.ONE_DAY_AWAY,
            QuickModes.SYSTEM_OFF,
        ):
            mode = ActiveMode(Zone.MIN_TARGET_HEATING_TEMP, quick_mode)

        if quick_mode == QuickModes.ONE_DAY_AT_HOME:
            if zone.heating:
                time_program = _get_setting_for_sunday(zone.heating.time_program)
                target_temp = zone.heating.target_high
                if time_program.setting == SettingModes.NIGHT:
                    target_temp = zone.heating.target_low
                mode = ActiveMode(target_temp, quick_mode)

        if quick_mode == QuickModes.PARTY:
            if zone.heating:
                mode = ActiveMode(zone.heating.target_high, quick_mode)

        if quick_mode == QuickModes.COOLING_FOR_X_DAYS and zone.cooling:
            mode = ActiveMode(zone.cooling.target_high, quick_mode)

    return mode


def _active_mode_for_room(
    room: Room, holiday: HolidayMode, quick_mode: QuickMode
) -> Optional[ActiveMode]:
    """Get the current
    :class:`~pymultimatic.model.mode.ActiveMode` for a
    :class:`~pymultimatic.model.component.Room`. This is the only way to
    get the real one.

    Args:
        room (Room): The room you want the active mode for.

    Returns:
        ActiveMode: The active mode.
    """
    # Holiday mode takes precedence over everything
    if holiday and holiday.active_mode:
        return holiday.active_mode

    # Global system quick mode takes over room settings
    if quick_mode and quick_mode.for_room:
        if quick_mode == QuickModes.SYSTEM_OFF:
            return ActiveMode(Room.MIN_TARGET_TEMP, quick_mode)

    return room.active_mode


def _active_mode_for_circulation(
    circulation: Circulation, holiday: HolidayMode, quick_mode: QuickMode
) -> Optional[ActiveMode]:
    """Get the current
    :class:`~pymultimatic.model.mode.ActiveMode` for a
    :class:`~pymultimatic.model.component.Circulation`. This is the only
    way to get the real one.

    Args:
        circulation (Circulation): The circulation you want the active mode
        for. If ``None`` is passed, it will pick up the circulation from
        the system.

    Returns:
        ActiveMode: The active mode.
    """
    if holiday and holiday.active_mode:
        active_mode = holiday.active_mode
        active_mode.target = None
        return active_mode

    if quick_mode and quick_mode.for_dhw:
        return ActiveMode(None, quick_mode)

    return circulation.active_mode


def _active_mode_for_hot_water(
    hot_water: HotWater, holiday: HolidayMode, quick_mode: QuickMode
) -> Optional[ActiveMode]:
    """Get the current
    :class:`~pymultimatic.model.mode.ActiveMode` for a
    :class:`~pymultimatic.model.component.HotWater`. This is the only way
    to get the real one.

    Args:
        hot_water (HotWater): The hot water you want the active mode
        for. If ``None`` is passed, it will pick up the hot water from the
        system.

    Returns:
        ActiveMode: The active mode.
    """

    if holiday and holiday.active_mode:
        active_mode = holiday.active_mode
        active_mode.target = constants.FROST_PROTECTION_TEMP
        return active_mode

    if quick_mode and quick_mode.for_dhw:
        if quick_mode == QuickModes.HOTWATER_BOOST:
            return ActiveMode(hot_water.target_high, quick_mode)

        if quick_mode == QuickModes.SYSTEM_OFF:
            return ActiveMode(constants.FROST_PROTECTION_TEMP, quick_mode)

        if quick_mode == QuickModes.ONE_DAY_AWAY:
            return ActiveMode(constants.FROST_PROTECTION_TEMP, quick_mode)

        if quick_mode == QuickModes.PARTY:
            return ActiveMode(hot_water.target_high, quick_mode)

    return hot_water.active_mode


def _active_mode_for_ventilation(
    ventilation: Ventilation, holiday: HolidayMode, quick_mode: QuickMode
) -> Optional[ActiveMode]:
    """Get the current
    :class:`~pymultimatic.model.mode.ActiveMode` for a
    :class:`~pymultimatic.model.Ventilation`. This is the only way
    to get the real one.

    Returns:
        ActiveMode: The active mode.
    """
    mode: ActiveMode = ventilation.active_mode

    if holiday and holiday.active_mode:
        active_mode = holiday.active_mode
        active_mode.target = Ventilation.MIN_LEVEL
        return active_mode

    if quick_mode and quick_mode.for_ventilation:
        if quick_mode == QuickModes.VENTILATION_BOOST:
            mode = ActiveMode(Ventilation.MAX_LEVEL, quick_mode)
        elif quick_mode in (QuickModes.ONE_DAY_AWAY, QuickModes.SYSTEM_OFF):
            mode = ActiveMode(Ventilation.MIN_LEVEL, quick_mode)
        elif quick_mode == QuickModes.PARTY:
            return ActiveMode(ventilation.target_high, quick_mode)
    else:
        mode = ventilation.active_mode
    return mode


def _get_setting_for_sunday(time_program: TimeProgram) -> Optional[TimePeriodSetting]:
    today = datetime.now()
    sunday = today - timedelta(days=today.weekday() - 6)
    return time_program.get_for(sunday)


_method_by_component: Dict[str, Callable[..., Optional[ActiveMode]]] = {
    Zone.__name__: _active_mode_for_zone,
    Room.__name__: _active_mode_for_room,
    Circulation.__name__: _active_mode_for_circulation,
    HotWater.__name__: _active_mode_for_hot_water,
    Ventilation.__name__: _active_mode_for_ventilation,
}



"""Test for quick mode."""
import unittest

from pymultimatic.model import OperatingModes, QuickModes, SettingModes


class ModeTest(unittest.TestCase):
    def test_hashable(self) -> None:
        test = {
            OperatingModes.AUTO: "test",
            QuickModes.HOTWATER_BOOST: "test",
            SettingModes.ON: "test",
        }

        self.assertEqual(3, len(test))



"""Tests for boiler status."""
import unittest
from datetime import datetime

from pymultimatic.model import BoilerStatus


class BoilerStatusTest(unittest.TestCase):
    """Test class."""

    def test_status_error_con(self) -> None:
        """Error code 'con' means error."""
        status = BoilerStatus("Name", "title", "con", "desc", datetime.now(), "hint")
        self.assertTrue(status.is_error)

    def test_status_error_f(self) -> None:
        """Error code starting with F means error."""
        status = BoilerStatus("Name", "title", "F.28", "desc", datetime.now(), "hint")
        self.assertTrue(status.is_error)

    def test_status_no_error(self) -> None:
        """No error code."""
        status = BoilerStatus("Name", "title", "S.04", "desc", datetime.now(), "hint")
        self.assertFalse(status.is_error)

    def test_status_no_code(self) -> None:
        """No code available."""
        status = BoilerStatus("Name", "title", None, "desc", datetime.now(), "hint")
        self.assertFalse(status.is_error)
